/***********************************************************************
 *
 * File Name:
 *		PnpMain.c
 *
 * Description:
 *		Implementation of Palm Network Programming sample application's
 *      main form handler.
 *
 * Note:
 *      Because this is example code, most of the error checking and
 *      safety measures (such as setting unlocked pointers to NULL) 
 *      have been removed.  These measures would need to be added back
 *      for production-ready code.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpMain.h"
#include "PnpNet.h"           // PnpNetConnect(), PnpNetDisconnect()
#include "PnpLog.h"           // 
#include "PnpSample_Res.h"    // MainOpenButton, MainCloseButton
#include <SysEvtMgr.h>
#include "GForm.h"            // Greg's form-handling declarations.
#include "GLog.h"             // Greg's log-handling declarations.

// Private main form functions.
static Boolean    OnFrmOpen (EventPtr eventP);
static Boolean    OnFrmClose (EventPtr eventP);
static Boolean    OnCtlSelect (EventPtr eventP);
static Boolean    OnCtlRepeat (EventPtr eventP);
static Boolean    OnKeyDown (EventPtr eventP);
static Boolean    OnNilEvent (EventPtr eventP);
static void       PrvMainLookupName (void);
static void       PrvMainLookupAddress (void);
static void       PrvMainLookupService (void);
static Err        PrvMainOpenNetwork (void);
static Err        PrvMainCloseNetwork (void);
static void       PrvMainEnableControls (FormPtr formP);
static void       PrvMainLoadTable (FormPtr formP);
static void       PrvMainDrawItem(void* tableP, Int16 row, Int16 column, RectanglePtr bounds);
static void       PrvMainScroll (WinDirectionType direction, Boolean oneLine);
static void       PrvMainUpdateScrollers (void);
static void       PrvMainFirstTableRow (FormPtr formP);

// Private  main form constants.
const UInt16   kNoSelection = -1;
const UInt16   kLogNet = 0;

// Private main form variables.
static Char    sText [255];
static UInt16  sLogID = pnpLogNet;
static Boolean sDirtyFlag = false;
static Boolean sNetOpenFlag = false;

// Current form.
static FormPtr formP = NULL;

// Table range values.
static UInt16  sFirstVisible = kNoSelection;

/*
   MainRegister:
   Register the main form's event handlers.
   Parms:   (none)
   Return:  0        Success
            ! 0      Error
*/
Err
MainRegister (void)
{
   Err err = 0;

   // Clear any existing event handlers.
   err = GFrmClearEventHandlers (MainForm);
   err = GFrmSetEventHandler (MainForm, frmOpenEvent, OnFrmOpen);
   err = GFrmSetEventHandler (MainForm, frmCloseEvent, OnFrmClose);
   err = GFrmSetEventHandler (MainForm, ctlSelectEvent, OnCtlSelect);
   err = GFrmSetEventHandler (MainForm, ctlRepeatEvent, OnCtlRepeat);
   err = GFrmSetEventHandler (MainForm, keyDownEvent, OnKeyDown);
   err = GFrmSetEventHandler (MainForm, nilEvent, OnNilEvent);

   return err;
}

/*
   OnOpen:
   Respond to opening of main form.
   Params:  none
   Return:  none
*/
static Boolean
OnFrmOpen (EventPtr /*eventP*/)
{
   // Declare local variables.
   Err err = 0;
   TablePtr tableP = NULL;
   
   // Get the active form.
   FormPtr formP = FrmGetActiveForm ();

   // Initialize log.
   err = PnpLogInit ();

   // Initialize network access.
   err = PnpNetInit ();

   // Draw the form.
   FrmDrawForm (formP);
   
   // Set dirty flag.
   sDirtyFlag = true;
   
   // Event has been handled.
   return true;
}

/*
   OnClose:
   Respond to closing of main form.
   Params:  none
   Return:  none
*/
static Boolean
OnFrmClose (EventPtr /*eventP*/)
{
   // Declare local variables.
   Err err = 0;
   
   // Close the network, if its open.
   err = PrvMainCloseNetwork ();
   
   // Terminate network access.
   PnpNetTerm ();

   // Terminate log.
   PnpLogTerm ();

   // Set dirty flag.
   sDirtyFlag = true;
   
   // Return value.
   return true;
}

/*
   OnCtlSelect:
   Handle a command sent to the main form.
   Params:  buttonID    Identifier of selected button.
   Return:  true        Handled (successfully or not)
            false       Not handled
*/
static Boolean
OnCtlSelect (EventPtr eventP)
{
   // Declare local variables.
   Boolean handled = false;
   
   switch (eventP->data.ctlEnter.controlID)
   {
      case MainNameButton:
      {
      // User pressed "Name"

         // Lookup name.
         PrvMainLookupName ();

         // Invalidate log table first index.
         sFirstVisible = kNoSelection;

         // Set dirty flag.
         sDirtyFlag = true;
         
         // Set handled flag.
         handled = true;
         break;
      }

      case MainAddressButton:
      {
      // User pressed "Address"

         // Lookup name.
         PrvMainLookupAddress ();

         // Invalidate log table first index.
         sFirstVisible = kNoSelection;

         // Set dirty flag.
         sDirtyFlag = true;
         
         // Set handled flag.
         handled = true;
         break;
      }

      case MainServiceButton:
      {
      // User pressed "Service"

         // Lookup name.
         PrvMainLookupService ();

         // Invalidate log table first index.
         sFirstVisible = kNoSelection;

         // Set dirty flag.
         sDirtyFlag = true;
         
         // Set handled flag.
         handled = true;
         break;
      }

      default:
      {
         break;
      }
   }
   
   return handled;
}

Boolean
OnCtlRepeat (EventPtr eventP)
{
   // Check the button.
	if (eventP->data.ctlRepeat.controlID == MainScrollUpRepeating)
	{
		PrvMainScroll (winUp, true);

      // Set dirty flag.
      sDirtyFlag = true;
	}
	else if (eventP->data.ctlRepeat.controlID == MainScrollDownRepeating)
	{
		PrvMainScroll (winDown, true);

      // Set dirty flag.
      sDirtyFlag = true;
	}

	// Repeating controls don't repeat if handled is set true.
   return false;
}

Boolean
OnKeyDown (EventPtr eventP)
{
   // Declare local variables.
   Boolean handled = false;
   
   // Check the button.
	if (eventP->data.keyDown.chr == pageUpChr)
	{
		PrvMainScroll (winUp, true);

      // Set dirty flag.
      sDirtyFlag = true;

		handled = true;
	}
	else if (eventP->data.keyDown.chr == pageDownChr)
	{
		PrvMainScroll (winDown, true);

      // Set dirty flag.
      sDirtyFlag = true;

		handled = true;
	}
	
	// Return value.
	return handled;
}

static Boolean
OnNilEvent (EventPtr eventP)
{
   // Check if things have changed.
   if (sDirtyFlag)
   {
   // Things have changed.
   
      // Get the active form.
      FormPtr formP = FrmGetActiveForm ();
      
      // Go to first visible row on table.
      PrvMainFirstTableRow (formP);
      
      // Reload the table data.
      PrvMainLoadTable (formP);
      
      // Redraw the table.
      GFrmRedrawTable (formP, MainLogTable);
      
      // Selectively enable controls.
      PrvMainEnableControls (formP);
      
      // Clear "dirty" flag.
      sDirtyFlag = false;
   }

   // Return.   
   return true;
}

/*
   PrvMainLookupName:
   Handle the name button press event.  Puts up the "Name Lookup" dialog.
   If the user selects "OK", attempts to locate the IP address for the
   entered name.
   Params:  (none)
   Returns: (none)
*/
static void
PrvMainLookupName (void)
{
   // Declare local variables.
   FormPtr dialogP = NULL;
   NetIPAddr ipAddr = 0;
   UInt16 textSize = sizeof sText;

   // Load the dialog.
   dialogP = FrmInitForm (NameLookupForm);

   // Show the dialog, check if user selects okay.
   if (NameLookupOKButton == FrmDoDialog (dialogP))
   {
   // Lookup host name.
   
      // Make sure network is open.
      if (0 == PrvMainOpenNetwork ())
      {
      // Network is open.
      
         // Get host name.
         GFrmGetFieldText (dialogP, NameLookupNameField, sText, &textSize);

         // Get the ip address.
         ipAddr = PnpNetFindName (sText);
      }
   }

   // Unload the form.
   FrmDeleteForm (dialogP);
}

/*
   PrvMainLookupAddress:
   Handle the address button press event.  Puts up the "Address Lookup"
   dialog.  If the user selects "OK", attempts to locate the IP address
   for the entered address.
   Params:  (none)
   Returns: (none)
*/
static void
PrvMainLookupAddress (void)
{
   // Declare local variables.
   FormPtr dialogP = NULL;
   NetIPAddr ipAddr = 0;
   UInt16 textSize = sizeof sText;

   // Load the dialog.
   dialogP = FrmInitForm (AddressLookupForm);

   // Show the dialog, check if user selects okay.
   if (AddressLookupOKButton == FrmDoDialog (dialogP))
   {
   // Lookup host address.
   
      // Make sure network is open.
      if (0 == PrvMainOpenNetwork ())
      {
      // Network is open.
      
         // Get host address.
         GFrmGetFieldText (dialogP, AddressLookupAddressField, sText, &textSize);

         // Get the ip address.
         ipAddr = PnpNetFindAddress (sText);
      }
   }

   // Unload the form.
   FrmDeleteForm (dialogP);
}

/*
   PrvMainLookupService:
   Handle the service button press event.  Puts up the "Service Lookup"
   dialog.  If the user selects "OK", attempts to locate the IP port
   for the service.
   Params:  (none)
   Returns: (none)
*/
static void
PrvMainLookupService ()
{
   // Declare local variables.
   FormPtr dialogP = NULL;
   UInt16 protoNameSize = sizeof (sText) / 2;
   UInt16 servNameSize = protoNameSize;
   Char* protoNameP = sText;
   Char* servNameP = sText + protoNameSize;
   UInt16 port;

   // Load the dialog.
   dialogP = FrmInitForm (ServiceLookupForm);

   // Show the dialog, check if user selected okay.
   if (ServiceLookupOKButton == FrmDoDialog (dialogP))
   {
   // Lookup service.
   
      // Make sure network is open.
      if (0 == PrvMainOpenNetwork ())
      {
         // Retrieve selected protocol, service.
         GFrmGetListText (dialogP, ServiceLookupProtocolList, protoNameP, &protoNameSize);
         GFrmGetListText (dialogP, ServiceLookupServiceList, servNameP, &servNameSize);
         
         // Retrieve the port for the service requested.
         port = PnpNetFindService (protoNameP, servNameP);
      }
   }

   // Unload the form.
   FrmDeleteForm (dialogP);
}

/*
   PrvMainOpenNetwork:
   Check if the network is opened and open if it isn't.
   Params:  (none)
   Returns: 0        Network open.
            <0       Error opening network.
*/
static Err
PrvMainOpenNetwork (void)
{
   // Declare local variables.
   Err err = 0;

   // Check if network is open.
   if (!sNetOpenFlag)
   {
      // Confirm user wants to connect.
      if (NetworkConnectOK == FrmAlert (NetworkConnectAlert))
      {
      // User confirmed.

         // Network not open - open it.
         err = PnpNetConnect ();

         // Check if network opened successfully.
         sNetOpenFlag = (!err);
      }
   }

   // Return value.
   return err;
}

/*
   PrvMainCloseNetwork:
   Close the network, if it's open.
   Params:  (none)
   Returns: 0        Network closed.
            <0       Error closing network.
*/
static Err
PrvMainCloseNetwork (void)
{
   // Declare local variables.
   Err err = 0;

   // Check if network is open.
   if (sNetOpenFlag)
   {
      // Network is open - close it.
      err = PnpNetDisconnect ();

      // Set network open flag to false.
      sNetOpenFlag = false;
   }

   // Return value.
   return err;
}

/*
   PrvMainEnableControls:
   Selectively enable controls depending on state.
   Params:  (none)
   Returns: (none)
*/
static void
PrvMainEnableControls (FormPtr formP)
{
   // Return.
   return;
}

static void
PrvMainLoadTable (FormPtr formP)
{
   // Declare local variables.
   UInt16 itemIndex = sFirstVisible;
   TablePtr tableP = NULL;
   UInt16 rowsInTable, itemsInLog, row;

   tableP = GFrmGetTable (formP, MainLogTable);
   rowsInTable = TblGetNumberOfRows (tableP);

   // Get the number of items in the log.
   itemsInLog = PnpLogNumItems (sLogID);

   for (row = 0; row < rowsInTable; row++, itemIndex++)
   {
		// Check if row is in range of log items.
      if (itemIndex < itemsInLog)
      {
			// Set customTableItem to indicate that we want to be called
			// to draw the record.  The callback set in PrvMainLoadRecords
			// will be called to draw the record.
			TblSetItemStyle (tableP, row, 0, customTableItem);
			TblSetRowUsable (tableP, row, true);
		}
		else
		{
			// Sometimes there are more table rows than records.
			// If this happens mark those unused table rows as
			// not usable.
			TblSetRowUsable (tableP, row, false);
		}

		// Mark the row invalid so that it will draw when we call the 
		// draw routine.
		TblMarkRowInvalid (tableP, row);
   }

   // Update the scroll arrows.
   PrvMainUpdateScrollers ();

	// Set the callback routine that will draw the records.
	TblSetCustomDrawProcedure (tableP, 0, PrvMainDrawItem);

	// Set the column usable so that it draws and accepts user input.
	TblSetColumnUsable (tableP, 0, true);	
}

static void
PrvMainDrawItem(void* tableVoidP, Int16 row, Int16 column, RectanglePtr bounds)
{
   TablePtr tableP = (TablePtr) tableVoidP;
	FontID currFont;
   UInt16 itemIndex;
   Char* textP = NULL;
   UInt16 size;
	Int16 textLen, width;
	Boolean fits;
	
	// Set the standard font.  Save the current font.
	// It is a Pilot convention to not destroy the current font
	// but to save and restore it.
	currFont = FntSetFont (stdFont);

	// Get the record number that corresponds to the table item to draw.
   // Calculate item index.
   itemIndex = sFirstVisible + row;

	// Retrieve size of log item.
   PnpLogGetItem (sLogID, itemIndex, NULL, &size);

   // Allocate buffer.
   textP = MemPtrNew (++size);

   // Check if allocation succeeded.
   if (textP)
   {
   // Allocation succeeded.

      // Retrieve block.
      PnpLogGetItem (sLogID, itemIndex, textP, &size);

   	// Determine the length of text that will fit within the bounds.
   	width = bounds->extent.x - 2;
   	textLen = StrLen(textP);
   	FntCharsInWidth (textP, &width, &textLen, &fits);
   	
   	// Now draw the text from the record.
   	WinDrawChars(textP, textLen, bounds->topLeft.x, bounds->topLeft.y);
	
      // Free the memory.
      MemPtrFree (textP);
   }

	// Restore the font.
	FntSetFont (currFont);
}

static void
PrvMainScroll (WinDirectionType direction, Boolean oneLine)
{
   // Declare local constants.
   const UInt16 oldFirstVisible = sFirstVisible;

   // Declare local variables.
   FormPtr formP = NULL;
	TablePtr tableP = NULL;
	UInt16 rowsInTable, itemsInLog;
	UInt16 newFirstVisible, newLastVisible;

   // Retrieve the active form.
   formP = FrmGetActiveForm ();

	// Get the table and it's number of rows
   tableP = GFrmGetTable (formP, MainLogTable);
	rowsInTable = TblGetNumberOfRows (tableP);

   // Get the number of items in the log.
   itemsInLog = PnpLogNumItems (sLogID);

	newFirstVisible = sFirstVisible;

	// Scroll the table down.
	if (direction == winDown)
	{
		// Scroll down a single line.
		if (oneLine)
      {
         // Calculate what the new last item would be.
         newFirstVisible = sFirstVisible + 1;
         newLastVisible = newFirstVisible + (rowsInTable - 1);

         // Check if we can scroll down one line.
         if (newLastVisible > itemsInLog)
         {
         // We can.

            // Increment first line.
            newFirstVisible = sFirstVisible;
			}
      }
		// Scroll down a page (less one row).
		else
		{
         // Calculate new last visible row index.
         newFirstVisible = sFirstVisible + (rowsInTable - 1);
         newLastVisible = newFirstVisible + (rowsInTable - 1);

         // Check if we can scroll down one page.
         if (newLastVisible > itemsInLog)
         {
            // Set last visible index to last item.
            newLastVisible = itemsInLog;

            // Calculate first item index.
            newFirstVisible = newLastVisible - (rowsInTable - 1);
         }
		}
   }
   else
   {
		// Scroll up a single line.
		if (oneLine)
      {
         // Check if we can scroll up.
         if (sFirstVisible > 0)
         {
            // Calculate new first item (one less).
            newFirstVisible = sFirstVisible - 1;
         }
      }
      else
      {
      // Try to scroll one page.
        
         // Check if we can scroll up.
         if (sFirstVisible > (rowsInTable-1))
         {
         // We can scroll up.

            // Scroll up one page.
            newFirstVisible = sFirstVisible - (rowsInTable - 1);
         }
         else
         {
         // Can't scroll up.

            // Set new first position to start of log.
            newFirstVisible = 0;
         }
      }
   }

   // Assign new first visible index to static.
   sFirstVisible = newFirstVisible;

	// Avoid redraw if no change
	if (sFirstVisible != oldFirstVisible)
	{
	// First visible has changed, redraw the table.

      // Get the active form.
      FormPtr formP = FrmGetActiveForm ();

      // Reload table.
      PrvMainLoadTable (formP);
            
      // Redraw the table.
      GFrmRedrawTable (formP, MainLogTable);
	}
}

static void
PrvMainUpdateScrollers (void)
{
   // Declare local variables.
   FormPtr formP = NULL;
	TablePtr tableP;
	UInt16 rowsInTable, itemsInLog;
	Boolean scrollUp, scrollDown;
	UInt16 upIndex, downIndex;

   // Retrieve the active form.
   formP = FrmGetActiveForm ();

	// Get the table and it's number of rows
   tableP = GFrmGetTable (formP, MainLogTable);
	rowsInTable = TblGetNumberOfRows (tableP);

   // Get the number of items in the log.
   itemsInLog = PnpLogNumItems (sLogID);
		
	// If the first item displayed is not the first item in the log
	// enable the up scroller.
	scrollUp = (sFirstVisible > 0);

	// If the last item displayed is not the last item in the log
	// enable the down scroller.
	scrollDown = (sFirstVisible + (rowsInTable) < itemsInLog);

	// Update the scroll button.
	upIndex = FrmGetObjectIndex (formP, MainScrollUpRepeating);
	downIndex = FrmGetObjectIndex (formP, MainScrollDownRepeating);
	FrmUpdateScrollers (formP, upIndex, downIndex, scrollUp, scrollDown);
}

static void
PrvMainFirstTableRow (FormPtr formP)
{
   // Declare local variables.
   TablePtr tableP;
   UInt16 rowsInTable, itemsInLog;

   if (kNoSelection == sFirstVisible)
   {   
      // Get the number of rows in the table.
      tableP = GFrmGetTable (formP, MainLogTable);
      rowsInTable = TblGetNumberOfRows (tableP);

      // Get the number of items in the log.
      itemsInLog = PnpLogNumItems (sLogID);

      // Check if there are more items than rows.
      if (itemsInLog > rowsInTable)
      {
         sFirstVisible = itemsInLog - rowsInTable;
      }
      else
      {
      // Fewer or equal items than rows.

         // Set first visible to first item.
         sFirstVisible = 0;
      }
   }
}
