/***********************************************************************
 *
 * FileName:
 *		PnpNet.c
 *
 * Description:
 *    Implementation of network access functions for "Palm Network
 *    Programming" sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 *    All network error checking code has been retained for purposes of
 *    example.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpNet.h"     // PnpNetConnect(), PnpNetDisconnect().
#include "PnpLog.h"     // PnpLogAddCaption(), PnpLogAddText(),
                        //    PnpLogAddResources().
#include <NetMgr.h>     // NetLib API.

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
void PrvNetLogOpenCount (void);
void PrvNetLogInterfaces (void);
void PrvNetLogAliases (Char** nameAliasesP);
void PrvNetLogAddresses (UInt8** addrListP, UInt16 addrType, UInt16 addrLen);

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
const UInt32      kNetTimeout = 10;
const NetIPAddr   kNilAddr = (NetIPAddr) 0;
const UInt16      kNilPort = (UInt16) 0;

// --------------------------------------------------------------------
// Private variable declarations
// --------------------------------------------------------------------
static UInt16 sNetLibRefNum = 0;            // NetLib reference number.
static UInt16 sOpenCount = 0;               // Local open count.
static UInt16 sState = pnpNetStateUninit;   // Network subsystem state.
static Char sText [255] = "";

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpNetState:
   Get the current state of the Network subsystem.
   Params:  (none)
   Returns: Current state of network subsystem.
*/
UInt16
PnpNetState (void)
{
   return sState;
}

/*
   PnpNetInit:
   Initialize the Network subsystem.  Checks for the
   existance and version of the net library, then
   obtains a reference number to it.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetInit ()
{
   // Declare local variables.
   Err err = 0;            // Error code.
   UInt32 version = 0L;    // NetLib version.
   UInt16 major = 0;       // Major version.
   UInt16 minor = 0;       // Minor version.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Initializing Network Library");

   // Check the ROM version.
   FtrGet (sysFtrCreator, sysFtrNumROMVersion, &version);

   // Extract the major, minor ROM versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Log ROM version.
   StrPrintF (sText, "ROM version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, sText);
   
   // Get the NetLib version.
   err = FtrGet (netFtrCreator, netFtrNumVersion, &version);

   // Check if NetLib version retrieved.
   if (0 != err)
   {
   // NetLib not installed.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not available.");

      // Return error.
      return err;
   }

   // Extract the major, minor NetLib versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Tell user the NetLib version.
   StrPrintF (sText, "NetLib version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, sText);

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib found");

   // Find the net library.
   err = SysLibFind ("Net.lib", &sNetLibRefNum);

   // Check if net library located.
   if (0 != err)
   {
   // Library not found.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not found.");

      // Return error.
      return err;
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib found");

   // Update state, we are now initialized.
   sState = pnpNetStateInit;

   // Return.
   return err;
}

/*
   PnpNetConnect:
   Connect to the network
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetConnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.
   UInt16 ifErrs = 0;     // Interface errors.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Connecting to Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib open");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Open the Net Library.
   err = NetLibOpen (sNetLibRefNum, &ifErrs);

   // Check if no errors.
   if (!err)
   {
   // Open succeeded.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib open");
   }

   // Check if the error is that we're already open.
   else if (netErrAlreadyOpen == err)
   {
   // NetLib is already open.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib already open");
   }

   else
   {
   // Open failed.

      // Draw current state.
      PnpLogAddItem (pnpLogNet, "Error opening NetLib.");

      // Return the error.
      return err;
   }

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Check if any interface errors occurred.
   if (ifErrs)
   {
   // One or more interface errors occurred.

      // Inform user of interface errors.
      PnpLogAddItem (pnpLogNet, "One or more interface errors");
   }

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();

   // Increment local open count.
   ++sOpenCount;

   // Update state, we are now connected.
   sState = pnpNetStateConnect;

   // Return success.
   return 0;
}

/*
   PnpNetDisconnect:
   Disconnect from the network
   Parms:   (none)
   Returns: 0        Success.
*/
Err
PnpNetDisconnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Disconnecting from Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib close");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Close the Net Library.
   err = NetLibClose (sNetLibRefNum, false);

   // Check if library was not open.
   if (netErrNotOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib was not open.");

      // Return.
      return 0;
   }

   // Check if we're still open.
   else if (netErrStillOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib is still open.");
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib close pending");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Now close the net lib.
   NetLibFinishCloseWait (sNetLibRefNum);

   // Draw current state.
   PnpLogAddResources (pnpLogNet, "NetLib closed");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();
   
   // Decrement local open count.
   --sOpenCount;

   // Return error.
   return 0;
}

/*
   PnpNetTerm:
   Terminate the Network subsystem.  Checks if the
   net library is still open.  If it is, closes it,
   but with a delay.
   Parms:   (none)
   Returns: 0        Success.
   Note:    This function contains no user interface
            code.
*/
Err
PnpNetTerm ()
{
   // Declare local variables.

   // Close any remaining opens.
   while (sOpenCount)
   {
      // Close net library.
      NetLibClose (sNetLibRefNum, false);

      // Decrement open count.
      --sOpenCount;
   }

   // Return success.
   return 0;
}

/*
   PnpNetFindName:
   Retrieve the ip address for a host name.
   Params:  hostNameP      Host name to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr
PnpNetFindName (Char* hostNameP)
{
   // Declare local variables.
   Err err = 0;
   NetHostInfoBufPtr hostInfoBufP = NULL;
   NetHostInfoPtr hostInfoP = NULL;
   NetIPAddr addr = kNilAddr;
   NetIPAddr* addrP = NULL;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Name Lookup");

   // Output name we are looking for.
   PnpLogAddItem (pnpLogNet, "Request:");
   StrPrintF (sText, "   %s", hostNameP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate host info buffer.
   hostInfoBufP = 
      (NetHostInfoBufPtr) MemPtrNew (sizeof (NetHostInfoBufType));

   // Check if allocation succeeded.
   if (!hostInfoBufP)
   {
      // Allocation failed, return error.
      return kNilAddr;
   }

   // Get the host information.
   hostInfoP = NetLibGetHostByName (sNetLibRefNum,
                                    hostNameP,
                                    hostInfoBufP,
                                    kNetTimeout * SysTicksPerSecond (),
                                    &err);

   // Check if host name retrieved.
   if (hostInfoP)
   {
      // Get pointer to first address.
      // (This is the one we'll return).
      addrP = (NetIPAddr*) hostInfoP->addrListP [0];

      // Set return value.
      // (Convert from network to host byte order).
      addr = NetNToHL (*addrP);

      // Inform user of results.
      PnpLogAddItem (pnpLogNet, "Found:");
      StrPrintF (sText, "   %s", hostInfoP->nameP);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (hostInfoP->nameAliasesP);

      // Output addresses.
      PrvNetLogAddresses (hostInfoP->addrListP,
                          hostInfoP->addrType,
                          hostInfoP->addrLen);
   }
   else
   {
      // Error retrieving name, inform user.
      StrPrintF (sText, "Error %i retrieving host name", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free host info buffer.
   MemPtrFree (hostInfoBufP);

   // Return error.
   return addr;
}

/*
   PnpNetFindAddress:
   Retrieve the ip address for a host name.
   Params:  hostAddrP      Host address to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr
PnpNetFindAddress (Char* hostAddrP)
{
   // Declare local variables.
   Err err = 0;
   NetHostInfoBufPtr hostInfoBufP = NULL;
   NetHostInfoPtr hostInfoP = NULL;
   NetIPAddr addr = kNilAddr;
   NetIPAddr* addrP = NULL;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Address Lookup");

   // Output address we are looking for.
   PnpLogAddItem (pnpLogNet, "Request:");
   StrPrintF (sText, "   %s", hostAddrP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate host info buffer.
   hostInfoBufP = 
      (NetHostInfoBufPtr) MemPtrNew (sizeof (NetHostInfoBufType));

   // Check if allocation succeeded.
   if (!hostInfoBufP)
   {
      // Allocation failed, return error.
      return kNilAddr;
   }

   // Convert address string to address.
   addr = NetLibAddrAToIN (sNetLibRefNum, hostAddrP);

   // Get the host information.
   hostInfoP = NetLibGetHostByAddr (sNetLibRefNum,
                            (UInt8*) &addr,
                            sizeof (addr),
                            netSocketAddrINET,
                            hostInfoBufP,
                            kNetTimeout * SysTicksPerSecond (),
                            &err);

   // Check if host address retrieved.
   if (hostInfoP)
   {
      // Get pointer to first address.
      // (This is the one we'll return).
      addrP = (NetIPAddr*) hostInfoP->addrListP [0];

      // Set return value.
      // (Convert from network to host byte order).
      addr = NetNToHL (*addrP);

      // Inform user of results.
      PnpLogAddItem (pnpLogNet, "Found:");
      StrPrintF (sText, "   %s", hostInfoP->nameP);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (hostInfoP->nameAliasesP);

      // Output addresses.
      PrvNetLogAddresses (hostInfoP->addrListP,
                          hostInfoP->addrType,
                          hostInfoP->addrLen);

   }
   else
   {
      // Error retrieving address, inform user.
      StrPrintF (sText, "Error %i retrieving host address", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free host info buffer.
   MemPtrFree (hostInfoBufP);

   // Return error.
   return addr;
}

/*
   PnpNetFindService:
   Retrieve the port number for a standard TCP/IP service.
   Params:  protoNameP     Protocol to use (TCP or UDP).
            servNameP      Service to retrieve.
   Returns: > 0            Port number.
            < 0            Error.
*/
UInt16
PnpNetFindService (Char* protoNameP, Char* servNameP)
{
   // Declare local variables.
   Err err = 0;
   NetServInfoBufPtr servInfoBufP = NULL;
   NetServInfoPtr servInfoP = NULL;
   UInt16 port = kNilPort;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Service Lookup");

   // Output protocols we are looking for.
   StrPrintF (sText, "Request: %s over %s", servNameP, protoNameP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate service info buffer.
   servInfoBufP = 
      (NetServInfoBufPtr) MemPtrNew (sizeof (NetServInfoBufType));

   // Check if allocation succeeded.
   if (!servInfoBufP)
   {
      // Allocation failed, return error.
      return kNilPort;
   }

   // Get the service information.
   servInfoP = NetLibGetServByName (sNetLibRefNum,
                            servNameP,
                            protoNameP,
                            servInfoBufP,
                            kNetTimeout * SysTicksPerSecond (),
                            &err);

   // Check if service retrieved.
   if (servInfoP)
   {

      // Service retrieved, set return value
      // (Convert from network to host byte order).
      port = NetNToHS (servInfoP->port);

      // Output results. 
      StrPrintF (sText, "Found: %s on port %i",
                 servInfoP->nameP, servInfoP->port);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (servInfoP->nameAliasesP);
   }
   else
   {
      // Error retrieving service, inform user.
      StrPrintF (sText, "Error %i retrieving service", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free service name buffer.
   MemPtrFree (servInfoBufP);

   // Return error.
   return port;
}

// ---------------------------------------------------------------------------
// Private Pnp Network functions.
// ---------------------------------------------------------------------------

/*
   PrvNetLogOpenCount:
   Show the Net Library's open count.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogOpenCount (void)
{
   // Declare local variables.
   UInt16 openCount = 0;

   // Get the open count.
   NetLibOpenCount (sNetLibRefNum,     // Net Lib ref num.
                    &openCount);       // Target for count.

   // Format message.
   StrPrintF (sText, "Open Count: %i", openCount);

   // Show message.
   PnpLogAddItem (pnpLogNet, sText);

   // Return.
   return;
}

/*
   PrvNetLogInterfaces:
   Show the state of each attached interface.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogInterfaces ()
{
   // Declare local variables.
   Err err = 0;            // Return value.
   UInt16 index = 0;       // Iterator.
   UInt32 ifCreator = 0L;  // Interface creator id.
   UInt16 ifInstance = 0;  // Interface instance.
   UInt8 ifUp = false;     // Interface up/down flag.
   Char ifName [32] = "";  // Interface name buffer.
   UInt16 settingSize = 0; // Size of requested setting.
   // Traverse list of interfaces.
   while (true)
   {
      // Get the interface.
      err = NetLibIFGet (sNetLibRefNum,   // NetLib ref num.
                         index,           // 0-based index of interface.
                         &ifCreator,      // Interface's creator.
                         &ifInstance);    // Interface's instance number.

      // Check for error.
      if (err)
      {
      // An error occurred.

         // Break out of for loop.
         break;
      }

      // Get the interface name.
      StrCopy (ifName, "<Unknown>");
      settingSize = sizeof (ifName);
      err = NetLibIFSettingGet (sNetLibRefNum,      // NetLib ref num.
                          ifCreator,          // Interface's creator.
                          ifInstance,         // Interface's instance number.
                          netIFSettingName,   // Setting to retrieve.
                          ifName,             // Target for setting.
                          &settingSize);      // Size of target for setting.

      // Check if interface is up.
      settingSize = sizeof (ifUp);
      err = NetLibIFSettingGet (sNetLibRefNum,   // NetLib ref num.
                          ifCreator,       // Interface's creator.
                          ifInstance,      // Interface's instance number.
                          netIFSettingUp,  // Setting to retrieve.
                          &ifUp,           // Target for setting.
                          &settingSize);   // Size of target for setting.

      // Format message.
      StrPrintF (sText, "%s - %s", ifName, ifUp ? "up" : "down");

      // Show message.
      PnpLogAddItem (pnpLogNet, sText);

      // Increment the index.
      index++;
   }

   // Update state, we are now disconnected.
   sState = pnpNetStateDisconnect;

   // Return.
   return;
}

/*
   PrvNetLogAliases:
   Log a list of aliases.
   Params:  nameAliasesP    List of alliases to draw.
   Returns: (none)
*/
void
PrvNetLogAliases (Char** nameAliasesP)
{
   // Declare Local Variables.
   Int16 i;    // Iterator.

   // Check if there are any aliases.
   if (NULL != nameAliasesP [0])
   {
      // Log subheading.
      PnpLogAddItem (pnpLogNet, "Aliases:");

      // Cycle through the alias list.
      for (i = 0; NULL != nameAliasesP [i]; i++)
      {
         // Output the alias name.
         StrPrintF (sText, "   %s", nameAliasesP [i]);
         PnpLogAddItem (pnpLogNet, sText);
      }
   }

   // Return.
   return;
}

/*
   PrvNetLogAddresses:
   Log a list of host addresses.
   Params:  addrListP   List of pointer to addresses.
            addrType    Type of addresses. Currently only 
                        netSocketAddrINET addresses are supported.
            addrLen     Size of individual addresses. Not used.
   Returns: (none)
*/
void
PrvNetLogAddresses (UInt8** addrListP, UInt16 addrType, UInt16 /*addrLen*/)
{
   // Declare Local Variables.
   NetIPAddr* addrP;          // IP address pointer.
   Char addrText [20] = "";   // For address conversion.
   Int16 i;    // Iterator.

   // Make sure it is a supported type.
   ErrFatalDisplayIf (netSocketAddrINET != addrType, 
                      "Only Internet Addresses are supported!");
                      
   // Check if there are any addresses.
   if (NULL != addrListP [0])
   {
      // Log subheading.
      PnpLogAddItem (pnpLogNet, "Addresses:");

      // Iterate through each pointer in the address list.
      for (i = 0; NULL != addrListP [i]; i++)
      {
         // Get ip address pointer.
         addrP = (NetIPAddr*) addrListP [i];

         // Convert host address to string.
         NetLibAddrINToA (sNetLibRefNum,
                          *addrP,
                          addrText);
         
         // Output the address.
         StrPrintF (sText, "   %s", addrText);
         PnpLogAddItem (pnpLogNet, sText);
      }
   }

   // Return.
   return;
}
