/***********************************************************************
 *
 * FileName:
 *		PnpFtp.c
 *
 * Description:
 *    Implementation of FTP functions for "Palm Network Programming"
 *    sample FTP Client application.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Include FTP Function declarations.
#include "PnpFtp.h"
#include "PnpNet.h"
#include "PnpLog.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const UInt16 kBufSize = 256;

static const Char  kCRLF [] = { 0x0D, 0x0A, 0x00 };

static Char * pnpFtpOpNames [] =
{
  "USER",
  "PASS",
  "QUIT"
};

static UInt16 const pnpFtpOpCount = sizeof (pnpFtpOpNames) / sizeof (pnpFtpOpNames [0]);

static const Char* kDfltUser = "anonymous";
static const Char* kDfltPass = "pnp@gregsprogrammingworks.com";
   
// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static Err   PrvFtpNetConnect (PnpFtpInfoPtr infoP);
static Err   PrvFtpNetDisconnect (PnpFtpInfoPtr infoP);
static Int16 PrvFtpQueueCommand (PnpFtpInfoPtr infoP, PnpFtpOpCode opCode, const Char* argP);
static Int16 PrvFtpReadResponse (PnpFtpInfoPtr infoP);
static Err   PrvFtpWriteControl (PnpFtpInfoPtr infoP);
static Err   PrvFtpReadControl (PnpFtpInfoPtr infoP);
static Err   PrvFtpAddLogItem (PnpFtpInfoPtr infoP, Char* logItem);
static Int16 PrvFtpFindCRLF (const UInt8* bufP, UInt16 size);

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpFtpInit:
   Initialize FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpFtpInit (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Initialize Network library.
   err = PnpNetInit ();
   if (err)
   {
      return -1;
   }

   // Initialize FTP session information.
   infoP->isNetConnect = false;
   infoP->ctrlSocket = kNilSocket;
   infoP->ctrlInBufH = GBufNew (kBufSize);
   infoP->ctrlOutBufH = GBufNew (kBufSize);

   // Check allocations.
   ErrFatalDisplayIf (NULL == infoP->ctrlInBufH, "Error creating inbound buffer.");
   ErrFatalDisplayIf (NULL == infoP->ctrlOutBufH, "Error creating outbound buffer.");

   // Log successful initialization.
   PrvFtpAddLogItem (infoP, "[FTP Session initialized.]");

   // Return result.
   return 0;
}

/*
   PnpFtpTerm:
   Terminate FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpFtpTerm (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Disconnect from the network.
   // (This will end the FTP session, if necessary)
   PrvFtpNetDisconnect (infoP);
   
   // Check if inbound buffer allocated.
   if (NULL != infoP->ctrlInBufH)
   {
   // Buffer is allocated.

      // Free buffer.
      GBufFree (infoP->ctrlInBufH);
      infoP->ctrlInBufH = NULL;
   }

   // Check if inbound buffer allocated.
   if (NULL != infoP->ctrlOutBufH)
   {
   // Buffer is allocated.

      // Free buffer.
      GBufFree (infoP->ctrlOutBufH);
      infoP->ctrlOutBufH = NULL;
   }

   // Terminate network.
   PnpNetTerm ();

   // Log successful initialization.
   PrvFtpAddLogItem (infoP, "[FTP Session terminated.]");

   // Return result.
   return err;
}

/*
   PnpFtpLogin:
   Login to the FTP server.
   Params:  infoP    FTP Session object.
            svrP     Name of server to login to.
            usrP     User name to login with, or NULL for anonymous.
            pwdP     Password for User account, or NULL for anonymous.
   Returns: 0        Success.
            < 0      Error.
   Note:    FTP Login Works as follows:

            1. Client Connects to Server on Port 21.
            2. Server sends Client a response.
            3. Client sends Server user name.
            4. Server sends Client a response.
            5. Client sends Server password.
            6. Server sends Client a response.

            Assuming that each of these responses indicated success
            (indicated by response from server starting with 2 or 3)
            we are now logged in.
*/
Err
PnpFtpLogin (PnpFtpInfoPtr infoP, Char* svrP, Char* usrP, Char* pwdP)
{
   // Declare local variables.
   Err err = 0;
   Int16 resCode;

   // Make sure we are connected to the network.
   PrvFtpNetConnect (infoP);
   
   // Attempt to connect with the server.
   infoP->ctrlSocket = PnpNetSocketConnect (svrP, "ftp", "tcp");

   // Check result.
   if (kNilSocket == infoP->ctrlSocket)
   {
   // Error opening socket.

      // Return error.
      return -1;
   }

   // Read prelim. response from server.
   do
   {
      resCode = PrvFtpReadResponse (infoP);
   }  while (resCode > 0);

// Send user command.

   // Make sure user name is valid.
   usrP = usrP ? usrP : (Char*) kDfltUser;
   
   // Send USER command.
   resCode = PrvFtpQueueCommand (infoP, kOpCodeCmdUser, usrP);

   // Read server response (ignore any remaining welcome messages).
   do
   {
      resCode = PrvFtpReadResponse (infoP);
   }  while (200 <= resCode && resCode < 300);

   // Check if USER command succeeded.
   if (300 <= resCode && 400 > resCode)
   {
   // USER command succeeded.

      // Make sure password is valid.
      pwdP = pwdP ? pwdP : (Char*) kDfltPass;

      // Send PASSWORD command.
      resCode = PrvFtpQueueCommand (infoP, kOpCodeCmdPass, pwdP);

      // Read server response.
      resCode = PrvFtpReadResponse (infoP);

      // Check if PASSWORD command succeeded.
      if (200 <= resCode && 300 > resCode)
      {
      // PASSWORD command succeeded.

         // Set return value to success.
         err = 0;
      }
      else if (resCode > 0)
      {
      // PASSWORD command failed.

         // Pass along return value.
         err = -resCode;
      }
      else
      {
      // Unable to read a response.
      
         // Result code holds error.
         err = resCode;
      }
   }
   else if (resCode > 0)
   {
   // USER command failed.

      // Pass along return value.
      err = -resCode;
   }
   else
   {
   // Unable to read a response.
   
      // Result code holds error.
      err = resCode;
   }

   // Check for errors.
   if (err)
   {
   // An error occurred.

      // Disconnect socket.
      PnpNetSocketDisconnect (infoP->ctrlSocket);
      infoP->ctrlSocket = kNilSocket;
   }

   // Return value.
   return err;
}

/*
   PnpFtpLogout:
   Logout from FTP Server.
   Params:  infoP    FTP Session object.
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpFtpLogout (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Int16 resCode;

   // Send the QUIT command.
   resCode = PrvFtpQueueCommand (infoP, kOpCodeCmdQuit, NULL);

   // Shutdown the socket (for sending).
   PnpNetSocketShutdown (infoP->ctrlSocket);

   // Read any remaining data.
   do
   {
      resCode = PrvFtpReadResponse (infoP);
   }  while (0 < resCode);

   // Close the socket.
   PnpNetSocketDisconnect (infoP->ctrlSocket);

   // Clear session control socket attribute.
   infoP->ctrlSocket = kNilSocket;

   // Return success.
   return 0;
}

/*
   PnpFtpIsLogin:
   Check if the we are logged in to the FTP server.
   Params:  infoP    FTP Session information.
   Returns: true     Logged in to FTP server.
            false    Not logged in to FTP server.
*/
Boolean
PnpFtpIsLogin (PnpFtpInfoPtr infoP)
{
   // Return true if we are connected, false if not.
   return (kNilSocket != infoP->ctrlSocket) ? true : false;
}

// --------------------------------------------------------------------
// Private function implementations
// --------------------------------------------------------------------

/*
   PrvFtpNetConnect:
   Connect to the network, if not already connected.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error.
*/
static Err
PrvFtpNetConnect (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;   // Return value.

   // Check if we are connected to the network.
   if (false == infoP->isNetConnect)
   {
   // We are not connected.
   
      // Connect to the network.
      err = PnpNetConnect ();
      if (!err)
      {
      // Successfully connected to network.

         // Set network connection flag to true.
         infoP->isNetConnect = true;
      }
   }
   
   // Return error.
   return err;
}

/*
   PrvFtpNetDisconnect:
   Disconnect from the network, if connected.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error.
*/
static Err
PrvFtpNetDisconnect (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;   // Return value.

   // Check if we are connected to the network.
   if (false != infoP->isNetConnect)
   {
   // We are connected.
   
      // Check if we're logged in.
      // (We can't be logged in if we're not connected)
      if (PnpFtpIsLogin (infoP))
      {
      // We're still logged in.

         // Logout.
         PnpFtpLogout (infoP);
      }

      // Disconnect from the network.
      err = PnpNetDisconnect ();

      // Clear network connection flag.
      infoP->isNetConnect = false;
   }
   
   // Return error.
   return err;
}

/*
   PrvFtpQueueCommand:
   Format and write a command to the FTP server.
   Params:  infoP    FTP Session information.
            opCode   Operation code of command to send.
            argP     Argument for command, NULL if none.
   Returns: 0        Success.
            <0       Error sending command.
*/
static Int16
PrvFtpQueueCommand (PnpFtpInfoPtr infoP, PnpFtpOpCode cmdOpCode, const Char* argP)
{
   // Declare local variables.
   UInt16 size;   // Return value.
   Err err = 0;
   Char* sendBufP = NULL;
   Char* cmdP = NULL;
   
   ErrFatalDisplayIf (cmdOpCode >= pnpFtpOpCount, "Invalid FTP Operation Code!");
   
   // Get command from array.
   cmdP = pnpFtpOpNames [cmdOpCode];
   
   // Calc. size of string.
   size = StrLen (cmdP) + sizeof (kCRLF); // sizeof includes nil-terminator.
   if (argP)
   {
   // There's an argument.
     
      // Add size of space and length of argument to size.
      size += StrLen (argP) + 1;
      
      // Allocate the string.
      sendBufP = (Char*) MemPtrNew (size);
      
      // Format the string.
      size = StrPrintF (sendBufP, "%s %s%s", cmdP, argP, kCRLF);
   }
   else
   {
   // There's no argument.
     
      // Allocate the string.
      sendBufP = (Char*) MemPtrNew (size);
      
      // Format the string.
      size = StrPrintF (sendBufP, "%s%s", cmdP, kCRLF);
   }

   // Append the command to the buffer.
   GBufAppend (infoP->ctrlOutBufH, (UInt8*) sendBufP, size);
   
   // Add command to log, too.
   PrvFtpAddLogItem (infoP, sendBufP);
   
   // Free the command.
   MemPtrFree (sendBufP), sendBufP = NULL;
   
   // Write the data.
   err = PrvFtpWriteControl (infoP);
   
   // Return result.
   return size;
}

/*
   PrvFtpReadResponse:
   Read a response from the FTP server.
   Params:  infoP    FTP Session information.
   Returns: >0       Server response code.
            <0       Error receiving response.
*/
static Int16
PrvFtpReadResponse (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   Int16 retVal = 0;
   UInt16 size = 0, len = -1;
   const UInt8* bufP = NULL;
   Char* responseBufP = NULL;

   // Read any incoming data into buffers.
   PrvFtpReadControl (infoP);
 
   // Get read pointer to buffer.
   bufP = GBufGetReadPtr (infoP->ctrlInBufH, &size);

   // Check if there's anything in the buffer.
   if (NULL == bufP || 0 == size)
   {
   // No data to read.
   
      // Return error.
      return -1;
   }
   
   // Look for CRLF.
   len = PrvFtpFindCRLF (bufP, size);
   
   // Release read pointer.
   GBufReleaseReadPtr (infoP->ctrlInBufH);
   
   // Check if we found anything.
   if (0 < len)
   {
   // We found something.
   
      // Get the code.
      size = len + sizeof kCRLF;   // Leave room for CR, LF and Nil-terminator.
      
      // Allocate a buffer.
      responseBufP = (Char*) MemPtrNew (size);
      
      // Make sure allocation succeeded.
      ErrFatalDisplayIf (NULL == responseBufP, "Memory allocation failed!");
      
      // Clear the buffer.
      MemSet (responseBufP, size, '\0');
      
      // Copy response to buffer.
      GBufRead (infoP->ctrlInBufH, (UInt8*) responseBufP, size-1);
      
      // Add response to log.
      PrvFtpAddLogItem (infoP, responseBufP);
      
      // Now check read numeric value of first three chars.
      responseBufP [3] = '\0';
      
      // Get response code.
      retVal = StrAToI (responseBufP);
      
      // Free the buffer.
      MemPtrFree (responseBufP);
   }
   
   // Return value.
   return retVal;
}

/*
   PrvFtpWriteControl:
   Write buffered commands to the control connection to the FTP server.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error sending command or receiving response.
*/
static Err
PrvFtpWriteControl(PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   UInt16 totalToSend = 0, totalSent = 0, sent = 0;
   const UInt8* bufP = NULL;
   
   // Get a read pointer into the buffer.
   bufP = GBufGetReadPtr (infoP->ctrlOutBufH, &totalToSend);
   
   while (totalToSend > 0 && !err)
   {
      // Assign to send size to send size.
      sent = totalToSend;
      
      // Write command.
      err = PnpNetSocketSend (infoP->ctrlSocket, (UInt8*)bufP, &sent);

      // Check for error.
      if (!err)
      {
         // Update buffer pointer.
         bufP += sent;
         
         // Update statistics.
         totalToSend -= sent;
         totalSent += sent;
      }
   }
   
   // Release the pointer.
   GBufReleaseReadPtr (infoP->ctrlOutBufH);
   
   // Check for success.
   if (totalSent > 0)
   {
   // Remove sent data from buffer.

      // Clear the buffer.
      GBufRead (infoP->ctrlOutBufH, NULL, totalSent);
      
      // Clear any errors.
      err = 0;
   }
   
   // Return value.
   return err;
}

/*
   PrvFtpReadControl:
   Read buffered responses from the control connection to the FTP server.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            <0       Error sending command or receiving response.
*/
static Err
PrvFtpReadControl (PnpFtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   UInt16 totalToRead = 0, totalRead = 0, read = 0;
   UInt8* bufP = NULL, * readP = NULL;
   
   // Figure out how big a buffer we need.
   totalToRead = GBufAvail (infoP->ctrlInBufH);
   
   // See if we can get anything.
   if (totalToRead == (UInt16) -1 || totalToRead == (UInt16) 0)
   {
   // Error or no room.
   
      // Return error.
      return -1;
   }
   
   // Allocate a buffer big enough for all of it.
   bufP = (UInt8*) MemPtrNew (totalToRead);
   
   // Clear the buffer.
   MemSet (bufP, totalToRead, '\0');
   
   // Initialize read pointer.
   readP = bufP;
   
   // Begin reading.
   do
   {
      // Set read amount to new value.
      read = totalToRead;
      
      // Receive initial response.
      err = PnpNetSocketReceive (infoP->ctrlSocket, readP, &read);

      // Check for error.
      if (!err)
      {
      // Success.
      
         // Check if anything was actually read.
         if (0 < read)
         {
         // Something was read.
         
            // Update buffer pointer.
            readP += read;
            
            // Update statistics.
            totalToRead -= read;
            totalRead += read;
         }
         else
         {
         // No data to read.
         
            // Set error.
            // (This will be cleared below if data has been received)
            err = -1;
         }
      }
   // Check if there is more to read or an error occurred.
   }  while (totalToRead > 0 && !err);
   
   // Check if anything read.
   if (0 < totalRead)
   {
   // Something was read.
   
      // Add this to buffer.
      GBufAppend (infoP->ctrlInBufH, bufP, totalRead);
   
      // Clear any errors.
      err = 0;   
   }
  
   // Don't forget to release the buffer.
   MemPtrFree (bufP);   bufP = NULL;
   
   // Return result.
   return err;
}

static Err
PrvFtpAddLogItem (PnpFtpInfoPtr infoP, Char* logItem)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   Char* bufP;
   Int16 len, itemLen;

   // First let's get the log item string length.
   itemLen = StrLen (logItem);
   
   // Now check if there's a CRLF before the end.
   len = PrvFtpFindCRLF ((UInt8*) logItem, itemLen);
   if (0 > len)
   {
   // There was no CRLF.
   
      // Set length to length of string.
      len = itemLen;
   }
   
   // Allocate a temp block of memory.
   bufP = MemPtrNew (len+1);

   // Check if allocation succeeded.
   if (!bufP)
   {
   // Memory allocation failed.

      // Return approp. error.
      return -1;
   }

   // Copy log message to buffer.
   StrNCopy (bufP, logItem, len);
   bufP [len] = '\0';

   // Add result to log.
   err = PnpLogAddItem (pnpLogFtp, bufP);

   // Free the memory.
   MemPtrFree (bufP);

   // Return error.
   return err;
}
   
static Int16
PrvFtpFindCRLF (const UInt8* bufP, UInt16 size)
{
   // Declare local variables.
   Int16 i, len = -1;
   
   // Traverse read pointer looking for first CRLF.
   for (i = 0; -1 == len && i < size-1; i++)
   {
      if (bufP [i] == 0x0D && bufP [i+1] == 0x0A)
      {
         // Set length to i.
         len = i;
      }
   }
   
   // Return value.
   return len;
}
