/***********************************************************************
 *
 * File Name:
 *		PnpApp.c
 *
 * Description:
 *		Implementation of the Palm Network Programming sample application
 *    functions.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpDefs.h"    // Standard application definitions.
#include "PnpMain.h"    // MainFormHandleEvent().
#include "PnpSample_Res.h"     // Resource identifiers.
#include <SysEvtMgr.h>
#include "GForm.h"

// Private application functions.
static Err     AppStart (void);
static void    AppStop (void);
static void    AppEventLoop (void);
static Boolean AppHandleEvent (EventPtr);
static void    AppLoadForm (UInt16 formId);

/*
   PilotMain:
   Main entry point into pilot application.
   Parms:   launchCode     Launch code.
            infoP          Struct associated w/ launch code.
            launchFlags    Extra info about launch.
*/
UInt32 
PilotMain (UInt16 launchCode, MemPtr /*cmdPBP*/, UInt16 launchFlags)
{
   Err   err;

   switch (launchCode)
   {
      case sysAppLaunchCmdNormalLaunch:
      {
         // Initialize application.
         err = AppStart ();
         if (!err)
         {
            // Show the main form.
            FrmGotoForm (MainForm);
            AppEventLoop ();
            AppStop ();
         }
         break;
      }
      default:
      {
         err = 0;
         break;
      }
   }
   return err;
}

/*
   AppStart:
   Initialize the application.
   Parms:   (none)
   Return:  0 if success, or error code.
*/
Err
AppStart ()
{
   return 0;
}

/*
   AppStop ():
   Save the current state of the application.
   Parms:   (none)
   Return:  (none)
*/
void
AppStop ()
{
   // Close all forms.
   FrmCloseAllForms ();
}
   
/*
   AppEventLoop:
   Process events coming to application.
   Parms:   (none)
   Return:  (none)
   Note:    Application is in this function for majority
            of its execution.
*/
void
AppEventLoop (void)
{
   // Local constants.
   const UInt16   kEventTimeout = SysTicksPerSecond ();

   // Local variables.
   EventType   event;
   UInt16 err;
   
   do
   {
      EvtGetEvent (&event, kEventTimeout);
      
      // Ask system to handle event.
      if (false == SysHandleEvent (&event))
      {
         // Ask Menu to handle event.
         if (false == MenuHandleEvent (0, &event, &err))
         {
            // Menu did not handle event.
            // Ask App (that is, this) to handle event.
            if (false == AppHandleEvent (&event))
            {
               // App did not handle event.
               // Send event to appropriate form.
               FrmDispatchEvent (&event);
            }  // end if (false == AppHandleEvent (&event))
         }     // end if (false == MenuHandleEvent (0, &event, &err))
      }        // end if (false == SysHandleEvent (&event))
   }
   while (event.eType != appStopEvent);
}

/*
   AppEventHandler:
   Handle application-level events.
   Parms:   eventP   Event to handle.
   Return:  true     Handled (successfully or not)
            false    Not handled
*/
Boolean
AppHandleEvent (EventPtr eventP)
{
   Boolean handled = false;
   
   switch (eventP->eType)
   {
      case frmLoadEvent:
      {
         // Load the form resource.
         UInt16 formID = eventP->data.frmLoad.formID;
         
         AppLoadForm (formID);
         handled = true;
         break;
      }

      default:
      {
         break;
      }
   }
   return handled;
}

/*
   AppLoadForm:
   Load a form.
   Parms:   formID   Identifier of form to be loaded.
   Return:  (none)
*/
void
AppLoadForm (UInt16 formID)
{
   // Initialize the form.
   FormPtr formP = FrmInitForm (formID);
   
   // Show the form.
   FrmSetActiveForm (formP);
   
   // Set the event handler for the form.
   //    The handler of the currently active form is called by
   //    FrmHandleEvent each time it receives an event.
   switch (formID)
   {
      case MainForm:
      {
         MainRegister ();
         FrmSetEventHandler (formP, GFrmHandleEvent);
         break;
      }

      default:
      {
         break;
      }
   }
}

