/***********************************************************************
 *
 * FileName:
 *		PnpFtp.h
 *
 * Description:
 *    Declaration of FTP functions for "Palm Network Programming"
 *    FTP Client sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

#ifndef __PNPFTP_H__
#define __PNPFTP_H__

// Standard application definitions
#include "PnpDefs.h"
#include "GBuf.h"
#include "GQueue.h"
#include "GData.h"
#include <NetMgr.h>

// --------------------------------------------------------------------
// Public constant declarations
// --------------------------------------------------------------------

// FTP Transaction codes. 
typedef enum
{
   kTranCodeInactive = -1,
   kTranCodeCmdUser = 0,
   kTranCodeCmdPass,
   kTranCodeCmdQuit,
   kTranCodeCmdPort,
   kTranCodeCmdList,
   kTranCodeCmdLast,
   kTranCodeSktConnect,
   kTranCodeSktDisconnect
} PnpFtpTranCode;

// FTP Transaction State codes.
typedef enum
{
   kTranStateInactive,
   kTranStateSending,
   kTranStateReceiving,
   kTranStateComplete
} PnpFtpTranState;

// NEW!
// FTP Data Socket State codes.
typedef enum
{
   kDataStateNoTransfer,   // No data transfer. 
   kDataStateInactive,     // Not started.
   kDataStateListening,    // Listening for server connection request.
   kDataStateReceiving,    // Receiving data from server.
   kDataStateComplete      // Finished receiving data from server.
} PnpFtpDataState;

// --------------------------------------------------------------------
// Public structure declarations
// --------------------------------------------------------------------
typedef struct {
   Boolean isNetConnect;      // Are we connected to network.
   
   // FTP Control Socket information.
   NetSocketAddrType ctrlAddr; // Control server address.
   NetSocketRef ctrlSocket;   // Control socket reference.
   GBufHandle ctrlInBufH;     // Buffer for receiving data.
   GBufHandle ctrlOutBufH;    // Buffer for sending data.

   // FTP Transaction information.
   GQueue tranQ;                // Transaction queue.
   PnpFtpTranCode tranCode;
   PnpFtpTranState tranState;
   Int16 tranReply;

   // FTP Data Socket information.
   NetSocketRef dataSocket;   // Data socket reference.
   PnpFtpDataState dataState; // Data transfer state.
   GDataType data;            // Data object.
   
   // File Descriptor Sets for checking socket's Read, Write states.
   NetFDSetType readFDs;      // Read file descriptor set.
   NetFDSetType writeFDs;     // Write file descriptor set.

} PnpFtpInfoType, * PnpFtpInfoPtr;

// --------------------------------------------------------------------
// Public function declarations
// --------------------------------------------------------------------

/*
   PnpFtpInit:
   Initialize FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err PnpFtpInit (PnpFtpInfoPtr infoP);

/*
   PnpFtpTerm:
   Terminate FTP library.
   Params:  infoP    Pointer to FTP Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err PnpFtpTerm (PnpFtpInfoPtr infoP);

/*
   PnpFtpProcess:
   Process any incoming or outgoing FTP events.
   Params:  infoP    FTP Session information.
   Returns: 0        One or more FTP transactions completed.
            <0       No progress made.
*/
Err PnpFtpProcess (PnpFtpInfoPtr infoP);

/*
   PnpFtpLogin:
   Login to the FTP server.
   Params:  infoP    FTP Session object.
            svrP     Name of server to login to.
            usrP     User name to login with, or NULL for anonymous.
            passP    Password for User account, or NULL for anonymous.
   Returns: 0        Success.
            < 0      Error.
*/
Err PnpFtpLogin (PnpFtpInfoPtr infoP, Char* svrP, Char* usrP, Char* passP);

/*
   PnpFtpLogout:
   Logout from FTP Server.
   Params:  infoP    FTP Session object.
   Returns: 0        Success.
            < 0      Error.
*/
Err PnpFtpLogout (PnpFtpInfoPtr infoP);

/*
   PnpFtpCanLogin:
   Check if the we can login to an FTP server.
   Params:  infoP    FTP Session information.
   Returns: true     We are in a valid state for login.
            false    We are not in a valid state for login.
*/
Boolean PnpFtpCanLogin (PnpFtpInfoPtr infoP);

/*
   PnpFtpCanLogout:
   Check if the we can login to an FTP server.
   Params:  infoP    FTP Session information.
   Returns: true     We are in a valid state for login.
            false    We are not in a valid state for login.
*/
Boolean PnpFtpCanLogout (PnpFtpInfoPtr infoP);

/*
   PnpFtpList:
   List the contents of the server's current directory.
   Params:  infoP    FTP Session information.
   Returns: 0        Success.
            < 0      Error.
*/
Err PnpFtpList (PnpFtpInfoPtr infoP);

#endif	// __PNPFTP_H__
