/***********************************************************************
 *
 * FileName:
 *		PnpNet.c
 *
 * Description:
 *    Implementation of network access functions for "Palm Network
 *    Programming" sample application.
 *
 * Note:
 *    Because this is example code, most of the error checking and safety
 *    measures (such as setting unlocked pointers to NULL) have been removed.
 *    This would need to be added back for production-ready code.
 *
 *    All network error checking code has been retained for purposes of
 *    example.
 *
 * Copyright  2000, Greg Winton
 * All rights reserved.
 *
 **********************************************************************/

#include "PnpNet.h"     // PnpNetConnect(), PnpNetDisconnect().
#include "PnpLog.h"     // PnpLogAddCaption(), PnpLogAddText(),
                        //    PnpLogAddResources().
#include <NetMgr.h>     // NetLib API.

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static void PrvNetLogOpenCount (void);
static void PrvNetLogInterfaces (void);
static void PrvNetLogAliases (Char** nameAliasesP);
static void PrvNetLogAddresses (UInt8** addrListP);
static NetSocketTypeEnum PrvNetCalcSocketType (Char* protocol);
static void PrvNetLogSocketInfo (NetSocketRef socket);
static Int16 PrvNetSocketDmResize (NetSocketRef socketRef, GDataPtr dataP);

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const UInt32     kNetTimeout = 5;
static const NetIPAddr  kNilAddr    = (NetIPAddr) 0;
static const UInt16     kNilPort    = (UInt16) 0;
static const UInt16     kNilProto   = (UInt16) 0;
static const UInt16     kQueueLen   = 1;        // Higher values are ignored
static const UInt16     kMinDataRead   = 1024;
static const UInt16     kMaxDataSize   = 0x7fff;

#define prvNetTimeout      (kNetTimeout * SysTicksPerSecond ())
#define prvNetWaitForever  (-1)

// --------------------------------------------------------------------
// Private variable declarations
// --------------------------------------------------------------------
static UInt16 sNetLibRefNum = 0;             // NetLib reference number.
static UInt16 sOpenCount = 0;                // Local open count.
static UInt16 sState = pnpNetStateUninit;    // Network subsystem state.
static Char sText [255] = "";

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpNetState:
   Get the current state of the Network subsystem.
   Params:  (none)
   Returns: Current state of network subsystem.
*/
UInt16
PnpNetState (void)
{
   return sState;
}

/*
   PnpNetInit:
   Initialize the Network subsystem.  Checks for the
   existance and version of the net library, then
   obtains a reference number to it.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetInit ()
{
   // Declare local variables.
   Err err = 0;            // Error code.
   UInt32 version = 0L;    // NetLib version.
   UInt16 major = 0;       // Major version.
   UInt16 minor = 0;       // Minor version.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Initializing Network Library");

   // Check the ROM version.
   FtrGet (sysFtrCreator, sysFtrNumROMVersion, &version);

   // Extract the major, minor ROM versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Log ROM version.
   StrPrintF (sText, "ROM version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, sText);
   
   // Get the NetLib version.
   err = FtrGet (netFtrCreator, netFtrNumVersion, &version);

   // Check if NetLib version retrieved.
   if (0 != err)
   {
   // NetLib not installed.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not available.");

      // Return error.
      return err;
   }

   // Extract the major, minor NetLib versions.
   major = sysGetROMVerMajor(version);
   minor = sysGetROMVerMinor(version);

   // Tell user the NetLib version.
   StrPrintF (sText, "NetLib version %i.%i", major, minor);
   PnpLogAddItem (pnpLogNet, sText);

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib found");

   // Find the net library.
   err = SysLibFind ("Net.lib", &sNetLibRefNum);

   // Check if net library located.
   if (0 != err)
   {
   // Library not found.

      // Show error.
      PnpLogAddItem (pnpLogNet, "NetLib not found.");

      // Return error.
      return err;
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib found");

   // Update state, we are now initialized.
   sState = pnpNetStateInit;

   // Return.
   return err;
}

/*
   PnpNetConnect:
   Connect to the network
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpNetConnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.
   UInt16 ifErrs = 0;     // Interface errors.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Connecting to Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib open");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Open the Net Library.
   err = NetLibOpen (sNetLibRefNum, &ifErrs);

   // Check if no errors.
   if (!err)
   {
   // Open succeeded.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib open");
   }

   // Check if the error is that we're already open.
   else if (netErrAlreadyOpen == err)
   {
   // NetLib is already open.

      // Show available resources.
      PnpLogAddResources (pnpLogNet, "NetLib already open");
   }

   else
   {
   // Open failed.

      // Draw current state.
      PnpLogAddItem (pnpLogNet, "Error opening NetLib.");

      // Return the error.
      return err;
   }

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Check if any interface errors occurred.
   if (ifErrs)
   {
   // One or more interface errors occurred.

      // Inform user of interface errors.
      PnpLogAddItem (pnpLogNet, "One or more interface errors");
   }

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();

   // Increment local open count.
   ++sOpenCount;

   // Update state, we are now connected.
   sState = pnpNetStateConnect;

   // Return success.
   return 0;
}

/*
   PnpNetDisconnect:
   Disconnect from the network
   Parms:   (none)
   Returns: 0        Success.
*/
Err
PnpNetDisconnect ()
{
   // Declare local variables.
   Err err = 0;         // Error code.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Disconnecting from Network");

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "Before NetLib close");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Close the Net Library.
   err = NetLibClose (sNetLibRefNum, false);

   // Check if library was not open.
   if (netErrNotOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib was not open.");

      // Return.
      return 0;
   }

   // Check if we're still open.
   else if (netErrStillOpen == err)
   {
   // NetLib is already open.

      // Inform the user.
      PnpLogAddItem (pnpLogNet, "NetLib is still open.");
   }

   // Show available resources.
   PnpLogAddResources (pnpLogNet, "NetLib close pending");

   // Draw the open count.
   PrvNetLogOpenCount ();

   // Show the interfaces and their states.
   PrvNetLogInterfaces ();
   
   // Decrement local open count.
   --sOpenCount;

   // Return error.
   return 0;
}

/*
   PnpNetTerm:
   Terminate the Network subsystem.  Checks if the
   net library is still open.  If it is, closes it,
   but with a delay.
   Parms:   (none)
   Returns: 0        Success.
   Note:    This function contains no user interface
            code.
*/
Err
PnpNetTerm ()
{
   // Declare local variables.

   // Close any remaining opens.
   while (sOpenCount)
   {
      // Close net library.
      NetLibClose (sNetLibRefNum, false);

      // Decrement open count.
      --sOpenCount;
   }

   // Return success.
   return 0;
}

/*
   PnpNetFindName:
   Retrieve the ip address for a host name.
   Params:  hostNameP      Host name to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr
PnpNetFindName (Char* hostNameP)
{
   // Declare local variables.
   Err err = 0;
   NetHostInfoBufPtr hostInfoBufP = NULL;
   NetHostInfoPtr hostInfoP = NULL;
   NetIPAddr addr = kNilAddr;
   NetIPAddr* addrP = NULL;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Name Lookup");

   // Output name we are looking for.
   PnpLogAddItem (pnpLogNet, "Request:");
   StrPrintF (sText, "   %s", hostNameP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate host info buffer.
   hostInfoBufP = 
      (NetHostInfoBufPtr) MemPtrNew (sizeof (NetHostInfoBufType));

   // Check if allocation succeeded.
   if (!hostInfoBufP)
   {
      // Allocation failed, return error.
      return kNilAddr;
   }

   // Get the host information.
   hostInfoP = NetLibGetHostByName (sNetLibRefNum,
                                    hostNameP,
                                    hostInfoBufP,
                                    prvNetTimeout,
                                    &err);

   // Check if host name retrieved.
   if (hostInfoP)
   {
      // Get pointer to first address.
      // (This is the one we'll return).
      addrP = (NetIPAddr*) hostInfoP->addrListP [0];

      // Set return value.
      // (Convert from network to host byte order).
      addr = NetNToHL (*addrP);

      // Inform user of results.
      PnpLogAddItem (pnpLogNet, "Found:");
      StrPrintF (sText, "   %s", hostInfoP->nameP);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (hostInfoP->nameAliasesP);

      // Output addresses.
      PrvNetLogAddresses (hostInfoP->addrListP);
   }
   else
   {
      // Error retrieving name, inform user.
      StrPrintF (sText, "Error %i retrieving host name", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free host info buffer.
   MemPtrFree (hostInfoBufP);

   // Return error.
   return addr;
}

/*
   PnpNetFindAddress:
   Retrieve the ip address for a host name.
   Params:  hostAddrP      Host address to lookup.
   Returns: > 0            Host ip address.
            0              Error finding host name.
*/
NetIPAddr
PnpNetFindAddress (Char* hostAddrP)
{
   // Declare local variables.
   Err err = 0;
   NetHostInfoBufPtr hostInfoBufP = NULL;
   NetHostInfoPtr hostInfoP = NULL;
   NetIPAddr addr = kNilAddr;
   NetIPAddr* addrP = NULL;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Address Lookup");

   // Output address we are looking for.
   PnpLogAddItem (pnpLogNet, "Request:");
   StrPrintF (sText, "   %s", hostAddrP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate host info buffer.
   hostInfoBufP = 
      (NetHostInfoBufPtr) MemPtrNew (sizeof (NetHostInfoBufType));

   // Check if allocation succeeded.
   if (!hostInfoBufP)
   {
      // Allocation failed, return error.
      return kNilAddr;
   }

   // Convert address string to address.
   addr = NetLibAddrAToIN (sNetLibRefNum, hostAddrP);

   // Get the host information.
   hostInfoP = NetLibGetHostByAddr (sNetLibRefNum,
                            (UInt8*) &addr,
                            sizeof (addr),
                            netSocketAddrINET,
                            hostInfoBufP,
                            prvNetTimeout,
                            &err);

   // Check if host address retrieved.
   if (hostInfoP)
   {
      // Get pointer to first address.
      // (This is the one we'll return).
      addrP = (NetIPAddr*) hostInfoP->addrListP [0];

      // Set return value.
      // (Convert from network to host byte order).
      addr = NetNToHL (*addrP);

      // Inform user of results.
      PnpLogAddItem (pnpLogNet, "Found:");
      StrPrintF (sText, "   %s", hostInfoP->nameP);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (hostInfoP->nameAliasesP);

      // Output addresses.
      PrvNetLogAddresses (hostInfoP->addrListP);
   }
   else
   {
      // Error retrieving address, inform user.
      StrPrintF (sText, "Error %i retrieving host address", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free host info buffer.
   MemPtrFree (hostInfoBufP);

   // Return error.
   return addr;
}

/*
   PnpNetFindService:
   Retrieve the port number for a standard TCP/IP service.
   Params:  protoNameP     Protocol to use (TCP or UDP).
            servNameP      Service to retrieve.
   Returns: > 0            Port number.
            < 0            Error.
*/
UInt16
PnpNetFindService (Char* protoNameP, Char* servNameP)
{
   // Declare local variables.
   Err err = 0;
   NetServInfoBufPtr servInfoBufP = NULL;
   NetServInfoPtr servInfoP = NULL;
   UInt16 port = kNilPort;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Service Lookup");

   // Output protocols we are looking for.
   StrPrintF (sText, "Request: %s over %s", servNameP, protoNameP);
   PnpLogAddItem (pnpLogNet, sText);

   // Allocate service info buffer.
   servInfoBufP = 
      (NetServInfoBufPtr) MemPtrNew (sizeof (NetServInfoBufType));

   // Check if allocation succeeded.
   if (!servInfoBufP)
   {
      // Allocation failed, return error.
      return kNilPort;
   }

   // Get the service information.
   servInfoP = NetLibGetServByName (sNetLibRefNum,
                            servNameP,
                            protoNameP,
                            servInfoBufP,
                            prvNetTimeout,
                            &err);

   // Check if service retrieved.
   if (servInfoP)
   {

      // Service retrieved, set return value
      // (Convert from network to host byte order).
      port = NetNToHS (servInfoP->port);

      // Output results. 
      StrPrintF (sText, "Found: %s on port %i",
                 servInfoP->nameP, servInfoP->port);
      PnpLogAddItem (pnpLogNet, sText);

      // Output aliases.
      PrvNetLogAliases (servInfoP->nameAliasesP);
   }
   else
   {
      // Error retrieving service, inform user.
      StrPrintF (sText, "Error %i retrieving service", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Free service name buffer.
   MemPtrFree (servInfoBufP);

   // Return error.
   return port;
}

/*
   PnpNetCalcAddress:
   Retrieve the network address for a host, service and protocol.
   Params:  addr        Out-param for address.
            hostNameP   Remote host name/address string.
            servNameP   Service name string.
            protoNameP  Protocol name string (TCP or UDP).
   Returns: 0           Success, socket is connected.
            <0          Error.
*/
Err
PnpNetCalcAddress (NetSocketAddrType* addrP, Char* host, Char* serv,  Char* proto)
{
   // Declare local variables.
   NetSocketAddrINType* inetAddrP = (NetSocketAddrINType*) addrP;
   NetIPAddr ipAddr;    // 32-bit IP address.
   UInt16 port = 0;     // Service port id.

   // Clear out-param.
   MemSet (addrP, sizeof (NetSocketAddrINType), '\0');
   
   // Retrieve ip address.
   ipAddr = PnpNetFindName (host);
   if (!ipAddr)
   {
   // Invalid host name.

      // Return error.
      return -1;
   }

   // Retrieve service.
   port = PnpNetFindService (proto, serv);
   if (!port)
   {
   // Invalid service name.

      // Return error.
      return -2;
   }

   // Populate address structure.
   inetAddrP->family = netSocketAddrINET;
   inetAddrP->port = NetHToNS (port);
   inetAddrP->addr = NetHToNL (ipAddr);

   // Return success.
   return 0;
}                       

/*
   PnpNetSocketOpen:
   Open a socket in non-blocking mode.
   Params:  socketType  Stream (TCP) or datagram (UDP).
   Returns: >= 0        Success.
            kNilSocket  Error.
*/
NetSocketRef PnpNetSocketOpen (NetSocketTypeEnum socketType)
{
   // Declare local variables.
   NetSocketRef socketRef = kNilSocket;   // Return value.
   Err err = 0;
   Boolean flag;
   Int16 res = 0;
   
   // Try to open the socket.
   socketRef = NetLibSocketOpen (sNetLibRefNum,
                                 netSocketAddrINET,
                                 socketType,
                                 kNilProto,
                                 prvNetTimeout, &err);

   // Insist on success.
   ErrFatalDisplayIf (kNilSocket == socketRef,
                      "Error opening socket.");

   // Output success and resources.
   PnpLogAddResources (pnpLogNet, "Opened");

   // Set socket to non-blocking mode.
   flag = true;
   res = NetLibSocketOptionSet (sNetLibRefNum,
                                socketRef,
                                netSocketOptLevelSocket,
                                netSocketOptSockNonBlocking,
                                &flag, sizeof (flag),
                                prvNetTimeout, &err);
   
   // Insist on success.
   ErrFatalDisplayIf (0 > res, "Error setting non-blocking  socket.");
      
   // Turn off Nagle algorithm (turn nodelay on).
   flag = true;
   res = NetLibSocketOptionSet (sNetLibRefNum,
                                socketRef,
                                netSocketOptLevelTCP,
                                netSocketOptTCPNoDelay,
                                &flag, sizeof (flag),
                                prvNetTimeout, &err);
   
   // Insist on success.
   ErrFatalDisplayIf (0 > res, "Error setting no delay socket.");
      
   // Return error.
   return socketRef;
}

/*
   PnpNetSocketClose:
   Close a socket.
   Params:  socketRef   Socket to close.
   Returns: 0           Success.
            !0          Error.
   Note:    After this function socketRef is invalid and should
            be set to kNilSocket accordingly.
*/
Err PnpNetSocketClose (NetSocketRef socketRef)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   
   // Close down the socket.
   NetLibSocketClose (sNetLibRefNum,
                      socketRef,
                      prvNetTimeout, &err);

   PnpLogAddResources (pnpLogNet, "After closed");

   // Return value.
   return err;
}

/*
   PnpNetSocketConnect:
   Connect a socket to a remote host.
   Params:  socketRef   Socket to connect.
            addrP       Remote host Internet (TCP or UDP) address.
   Returns: 0           Success, socket is connected.
            -1          Error.
*/
Err
PnpNetSocketConnect (NetSocketRef socketRef, NetSocketAddrType* addrP)
{
   // Declare local variables.
   Err err = 0;            // Out-param for socket function errors.
   Int16 res = 0;          // Result from socket functions.

   // Attempt to connect.
   res = NetLibSocketConnect (sNetLibRefNum, socketRef,
                              addrP, sizeof (NetSocketAddrType),
                              prvNetTimeout, &err);

   // Check result.
   if (!res || netErrSocketAlreadyConnected == err)
   {
   // Socket has already been connected.
   
      // In this topsy-turvy non-blocking world, 
      // this actually indicates success.
   
      // Draw socket information.
      PrvNetLogSocketInfo (socketRef);

      // Output success and resources.
      PnpLogAddResources (pnpLogNet, "Connected");

      // Set return value.
      err = 0;
   }
   else if (netErrWouldBlock == err)
   {
   // If this is the first time through, the Net Library initiates
   // the connection request. Otherwise, this indicates that the
   // connection is not yet ready; the 3-way handshake has not completed.
   
      // This is an error, but the socket is left open
      // I guess there's nothing to do.
      StrPrintF (sText, "Connection Pending...");
      PnpLogAddItem (pnpLogNet, sText);
   }
   else
   {
   // Error!

      // Error connecting socket, add log item.
      StrPrintF (sText, "Error %i connecting socket", err);
      PnpLogAddItem (pnpLogNet, sText);
   }

   // Return value.
   return err;
}

/*
   PnpNetSocketDisconnect:
   Close a socket connection.
   Params:  socketRef   Socket reference.
   Returns: 0           Success.
            < 0         Error.
*/
Err
PnpNetSocketDisconnect (NetSocketRef socketRef)
{
   // Declare local variables.
   Err err = 0;            // Return value.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "Disconnecting Socket");

   // Shut down socket.
   err = PnpNetSocketShutdown (socketRef);
   
   // Close the socket.
   err = PnpNetSocketClose (socketRef);
   
   // Return error.
   return err;
}

/*
   PnpNetSocketShutdown:
   Shutdown a socket connection.
   Params:  socketRef   Socket reference.
   Returns: 0           Success.
            < 0         Error.
*/
Err
PnpNetSocketShutdown (NetSocketRef socketRef)
{
   // Declare local variables.
   Err err = 0;            // Return value.

   // Show current resources.
   PnpLogAddResources (pnpLogNet, "Before shutdown");

   // Shut down the socket for send.
   NetLibSocketShutdown (sNetLibRefNum,
                         socketRef,
	                      netSocketDirOutput,
                         prvNetTimeout, &err);

   // Show current resources.
   PnpLogAddResources (pnpLogNet, "After shutdown");

   // Return error.
   return err;
}

/*
   PnpNetSocketReceive:
   Read a block of data from a socket.
   Params:  socketRef   Socket reference.
            dataP        Target buffer for received data.
            sizeP       [in] Size of target buffer.
                        [out] Amount of data received.
   Returns: 0           Success.
            < 0         Error.
*/
Err
PnpNetSocketReceive (NetSocketRef socketRef, UInt8* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;   // Return value.
   UInt16 read;
   UInt8* readP = NULL;

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "READ DATA");

   // Output initial statistics.
   StrPrintF (sText, "Receiving %i (max) bytes", *sizeP);
   PnpLogAddItem (pnpLogNet, sText);

   // Clear output-buffer.
   MemSet (textP, (*sizeP), '\0');

   // Receive data into buffers.
   read = NetLibReceive (sNetLibRefNum,      // libRefNum
                         socketRef,          // socket
                         textP,              // bufP
                         *sizeP,             // bufLen
                         0,                  // flags
                         NULL,               // toAddrP
                         0,                  // toAddrLen
                         prvNetTimeout,        // timeout
//                         0,                  // timeout
                         &err);              // errP

   // Check result of receive operation.
   if (0 < read)
   {
   // Success.
   
      // Set size out-value.
      (*sizeP) = read;
       
      // Add log item.  
      StrPrintF (sText, "Read %i bytes.", read);
      PnpLogAddItem (pnpLogNet, sText);
   }
   else if (0 == read)
   {
   // Connection has been closed.
   
      // Set size out-value.
      (*sizeP) = 0;
      
      // Add log item.
      PnpLogAddItem (pnpLogNet, "Read: No data available.");
   }
   else
   {
   // An error occurred.
   
      // Set size out-value.
      (*sizeP) = 0;
      
      // Add log item.
      StrPrintF (sText, "Error %i reading data.", err);
      PnpLogAddItem (pnpLogNet, sText);
   }
      
   // Return error.
   return err;
}

/*
   PnpNetSocketSend:
   Write a block of data on a socket.
   Params:  socketRef   Socket reference.
            dataP       Data to send.
            sizeP       [IN]  Amount of data to send.
                        [OUT] Amount of data sent.
   Returns: 0           Success.
            < 0         Error.
*/
Err
PnpNetSocketSend (NetSocketRef socketRef, UInt8* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;   // Return value.
   UInt16 toSend, sent;

   // Calc. total size to send, including cr-lf.
   toSend = (*sizeP);

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "SEND DATA");

   // Output initial statistics.
   StrPrintF (sText, "Sending %i bytes", toSend);

   // Send the data.
   sent = NetLibSend (sNetLibRefNum,      // libRefNum
                      socketRef,          // socket
                      textP,              // bufP
                      toSend,             // bufLen
                      0,                  // flags
                      NULL,               // toAddrP
                      0,                  // toAddrLen
                      prvNetTimeout,        // timeout
//                      0,                  // timeout
                      &err);              // errP

   // Check result of send operation.
   if (0 < sent)
   {
   // Success.
   
      // Set size out-value.
      (*sizeP) = sent;
       
      // Add log item.  
      StrPrintF (sText, "Wrote %i bytes.", sent);
      PnpLogAddItem (pnpLogNet, sText);
   }
   else if (0 == sent)
   {
   // Connection has been closed.
   
      // Set size out-value.
      (*sizeP) = sent;
      
      // Add log item.
      StrPrintF (sText, "Connection closed (error = %i).", err);
      PnpLogAddItem (pnpLogNet, sText);
   }
   else
   {
   // An error occurred.
   
      // Set size out-value.
      (*sizeP) = 0;
      
      // Add log item.
      StrPrintF (sText, "Error %i writing data.", err);
      PnpLogAddItem (pnpLogNet, sText);
   }
      
   // Return error.
   return err;
}

/*
   PnpNetSelect:
   Check for I/O on one or more sockets. This function auto-
   matically checks for user input events such as a pen tap
   or key press.
   Params:  readFDsP    Set of descriptors to check for input.
            writeFDsP   Set of descriptors to check for output.
   Return:  0           Success.
            <0          Error occurred.
*/
Err
PnpNetSelect (UInt16 width, NetFDSetType* readFDsP, NetFDSetType* writeFDsP)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   NetFDSetType exceptFDs;
   Int16 numFDs;

   // Initialize exception file descriptor set.
   netFDZero (&exceptFDs);

   // Set stdin file descriptor in read FD set.
   netFDSet (sysFileDescStdIn, readFDsP);

   // Calculate max width.
   if (width <= sysFileDescStdIn)
   {
      width = sysFileDescStdIn + 1;
   }
   
   // Wait for one of the descriptors to be ready.
   numFDs = NetLibSelect (sNetLibRefNum,
                          width,
                          readFDsP, 
                          writeFDsP,
                          &exceptFDs,
                          0, &err);

   // Return value.
   return err;
}

/*
   PnpNetSocketListen:
   Put the socket into passive, listening mode.
   Params:  (none)
   Returns: >= 0        Success, return value is socket ref.
            kNilSocket  Error.
*/
NetSocketRef
PnpNetSocketListen (void)
{
   // Declare local variables.
   NetSocketRef socketRef = kNilSocket;   // Return value.
   Err err = 0;                           // Used for returning errors.
   UInt16 resCode = 0;                    // Result from socket functions.
   NetSocketAddrINType inetAddr;          // Internet address variable.
   Int16 inetAddrLen;                     // Internet address size.
   UInt16 len = 0;                        // Length variable.

   // Clear the window.
   PnpLogAddCaption (pnpLogNet, "LISTEN");

   // Draw current resources.
   PnpLogAddResources (pnpLogNet, "Before listening on socket");

   // Try to open the socket.
   socketRef = PnpNetSocketOpen (netSocketTypeStream);

   // Insist on success.
   ErrFatalDisplayIf (kNilSocket == socketRef, "Error opening socket!");  

   // Output success and resources.
   PnpLogAddResources (pnpLogNet, "Opened");

   // Populate address structure.
   inetAddr.family = netSocketAddrINET;
   inetAddr.addr   = NetHToNL (netIPAddrLocal);
   inetAddr.port   = NetHToNS (0);
   inetAddrLen     = sizeof (inetAddr);
   
   // Bind socket to port.
   resCode = NetLibSocketBind (sNetLibRefNum,
                               socketRef,
                               (NetSocketAddrType*) &inetAddr,
                               inetAddrLen,
                               prvNetTimeout,
                               &err);

   // Insist on success.
   ErrFatalDisplayIf (0 != resCode, "Error binding socket!");  

   // Output success and resources.
   PnpLogAddResources (pnpLogNet, "Bound");

   // Attempt to connect.
   resCode = NetLibSocketListen (sNetLibRefNum,
                                 socketRef,
                                 kQueueLen,
                                 prvNetTimeout,
                                 &err);

   // Insist on success.
   ErrFatalDisplayIf (0 != resCode, "Error listening on socket!");  

   // Output success and resources.
   PnpLogAddResources (pnpLogNet, "Listening");

   // Return value.
   return socketRef;
}

/*
   PnpNetSocketAccept:
   Accept an incoming connection request.
   Params:  listenRef   Listening socket to accept connection from.
   Returns: >= 0        Success, return value is socket ref.
            kNilSocket  Error.
*/
NetSocketRef PnpNetSocketAccept (NetSocketRef listenRef)
{
   // Declare local variables.
   NetSocketRef acceptRef = NULL;   // Return value.
   NetSocketAddrINType inetAddr;    // Target for requestor's Internet address.
   Int16 size = sizeof (inetAddr);  // Size of Internet address target.
   Err err = 0;
   
   // Clear address.
   MemSet (&inetAddr, size, '\0');
   
   // Accept the incoming connection.
   acceptRef = NetLibSocketAccept (sNetLibRefNum,
                                   listenRef,
                                   (NetSocketAddrType*) &inetAddr,
                                   &size,
                                   prvNetTimeout,
                                   &err);
                                   
   // Return result.
   return acceptRef;
}

/*
   PnpNetSocketDmReceive:
   Receive data directly at end database record.
   Params:  socketRef   Socket to read data from.
            dataP       Database/record information.
   Returns: 0           Success.
            < 0         Error. No data, record is full or other.
*/
Err
PnpNetSocketDmReceive (NetSocketRef socketRef, GDataPtr dataP)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   Int16 readSize = 0;
   MemPtr recordP = NULL;
   
   // Insist that record be valid in every way.
   ErrFatalDisplayIf (NULL == dataP->recordH, "Record is not available!");
   ErrFatalDisplayIf (false == dataP->canEdit, "Record is read-only!");
   
   // Get read size and resize as necessary.
   readSize = PrvNetSocketDmResize (socketRef, dataP);
   
   // Insist read size is valid.
   ErrFatalDisplayIf (0 > readSize, "Unexpected data record error!");
   
   
   // Lock the record.
   recordP = MemHandleLock (dataP->recordH);
   
   // Read data.
   readSize = NetLibDmReceive (sNetLibRefNum,
                               socketRef,
                               recordP,
                               dataP->recordLength,
                               readSize,
                               0,
                               NULL,
                               NULL,
                               prvNetTimeout,
                               &err);
                           
   if (0 < readSize)
   {
   // Success.                        
                    
      // Update record length.
      dataP->recordLength += readSize;
   }
   
   // Unlock the record.
   MemPtrUnlock (recordP); recordP = NULL;
   
   // Return value.
   return err;
}

/*
   PnpNetSocketAddress:
   Get the socket's IP address in dotted decimal notation.
   Params:  addrS       Target for address string.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketAddress (NetSocketRef /*socketRef*/, Char* addrS)
{
   Err err = 0;
   UInt16 index, buflen, ifInstance;
   UInt32 ifCreator;
   NetIPAddr ipAddr = netIPAddrLocal;
   
   for (index = 0; netIPAddrLocal == ipAddr; index++) 
   {
      err = NetLibIFGet(sNetLibRefNum, index, &ifCreator, &ifInstance);
      if (err)
         break;

      // Check for loopback driver (address = 127.0.0.1).
      if (netIFCreatorLoop == ifCreator)
      {
         continue;
      }
      
      buflen = sizeof(ipAddr);
      NetLibIFSettingGet(sNetLibRefNum, ifCreator, ifInstance, netIFSettingActualIPAddr, &ipAddr, &buflen);
   }

   // Convert ip address to string.   
   NetLibAddrINToA (sNetLibRefNum, ipAddr, addrS);
   
   // Return error.
   return err;
}

/*
   PnpNetSocketPort:
   Get the socket's port number.
   Params:  socketRef   Socket reference.
            portNumP    Target for number.
   Returns: 0           Success.
            < 0         Error.
*/
Err PnpNetSocketPort (NetSocketRef socketRef, UInt16* portNumP)
{
   // Declare local variables.
   NetSocketAddrINType locAddr, remAddr;
   UInt16 locAddrLen, remAddrLen;
   Err err = 0;
   
   // Populate address structures.
   locAddr.family = remAddr.family = netSocketAddrINET;
   locAddr.addr   = remAddr.addr   = NetNToHS (netIPAddrLocal);
   locAddr.port   = remAddr.port   = NetHToNS (0);
   locAddrLen     = remAddrLen     = sizeof (NetSocketAddrINType);

   // Get the socket addresses.
   NetLibSocketAddr (sNetLibRefNum, socketRef,
                     (NetSocketAddrType*)&locAddr, (Int16*)&locAddrLen,
                     (NetSocketAddrType*)&remAddr, (Int16*)&remAddrLen,
                     prvNetTimeout, &err);

   // Set local port out-param - in host byte order.
   (*portNumP) = NetNToHS (locAddr.port);
   
   // Return result.
   return err;
}

// ---------------------------------------------------------------------------
// Private Pnp Network functions.
// ---------------------------------------------------------------------------

/*
   PrvNetLogOpenCount:
   Show the Net Library's open count.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogOpenCount (void)
{
   // Declare local variables.
   UInt16 openCount = 0;

   // Get the open count.
   NetLibOpenCount (sNetLibRefNum,     // Net Lib ref num.
                    &openCount);       // Target for count.

   // Format message.
   StrPrintF (sText, "Open Count: %i", openCount);

   // Show message.
   PnpLogAddItem (pnpLogNet, sText);

   // Return.
   return;
}

/*
   PrvNetLogInterfaces:
   Show the state of each attached interface.
   Parms:   (none)
   Returns: 0        Success.
            < 0      Error.
*/
void
PrvNetLogInterfaces ()
{
   // Declare local variables.
   Err err = 0;            // Return value.
   UInt16 index = 0;       // Iterator.
   UInt32 ifCreator = 0L;  // Interface creator id.
   UInt16 ifInstance = 0;  // Interface instance.
   UInt8 ifUp = false;     // Interface up/down flag.
   Char ifName [32] = "";  // Interface name buffer.
   UInt16 settingSize = 0; // Size of requested setting.
   
   // Traverse list of interfaces.
   while (true)
   {
      // Get the interface.
      err = NetLibIFGet (sNetLibRefNum,   // NetLib ref num.
                         index,           // 0-based index of interface.
                         &ifCreator,      // Interface's creator.
                         &ifInstance);    // Interface's instance number.

      // Check for error.
      if (err)
      {
      // An error occurred.

         // Break out of for loop.
         break;
      }

      // Get the interface name.
      StrCopy (ifName, "<Unknown>");
      settingSize = sizeof (ifName);
      err = NetLibIFSettingGet (sNetLibRefNum,      // NetLib ref num.
                          ifCreator,          // Interface's creator.
                          ifInstance,         // Interface's instance number.
                          netIFSettingName,   // Setting to retrieve.
                          ifName,             // Target for setting.
                          &settingSize);      // Size of target for setting.

      // Check if interface is up.
      settingSize = sizeof (ifUp);
      err = NetLibIFSettingGet (sNetLibRefNum,   // NetLib ref num.
                          ifCreator,       // Interface's creator.
                          ifInstance,      // Interface's instance number.
                          netIFSettingUp,  // Setting to retrieve.
                          &ifUp,           // Target for setting.
                          &settingSize);   // Size of target for setting.

      // Format message.
      StrPrintF (sText, "%s - %s", ifName, ifUp ? "up" : "down");

      // Show message.
      PnpLogAddItem (pnpLogNet, sText);

      // Increment the index.
      index++;
   }

   // Update state, we are now disconnected.
   sState = pnpNetStateDisconnect;

   // Return.
   return;
}

/*
   PrvNetLogAliases:
   Log a list of aliases.
   Params:  nameAliasesP    List of alliases to draw.
   Returns: (none)
*/
void
PrvNetLogAliases (Char** nameAliasesP)
{
   // Declare Local Variables.
   Int16 i;    // Iterator.

   // Check if there are any aliases.
   if (NULL != nameAliasesP [0])
   {
      // Log subheading.
      PnpLogAddItem (pnpLogNet, "Aliases:");

      // Cycle through the alias list.
      for (i = 0; NULL != nameAliasesP [i]; i++)
      {
         // Output the alias name.
         StrPrintF (sText, "   %s", nameAliasesP [i]);
         PnpLogAddItem (pnpLogNet, sText);
      }
   }

   // Return.
   return;
}

/*
   PrvNetLogAddresses:
   Log a list of host addresses.
   Params:  addrListP   List of pointer to addresses.
   Returns: (none)
*/
void
PrvNetLogAddresses (UInt8** addrListP)
{
   // Declare Local Variables.
   NetIPAddr* addrP;          // IP address pointer.
   Char addrText [20] = "";   // For address conversion.
   Int16 i;    // Iterator.

   // Check if there are any addresses.
   if (NULL != addrListP [0])
   {
      // Log subheading.
      PnpLogAddItem (pnpLogNet, "Addresses:");

      // Iterate through each pointer in the address list.
      for (i = 0; NULL != addrListP [i]; i++)
      {
         // Get ip address pointer.
         addrP = (NetIPAddr*) addrListP [i];

         // Convert host address to string.
         NetLibAddrINToA (sNetLibRefNum,
                          *addrP,
                          addrText);
         
         // Output the address.
         StrPrintF (sText, "   %s", addrText);
         PnpLogAddItem (pnpLogNet, sText);
      }
   }

   // Return.
   return;
}

/*
   PrvNetCalcSocketType:
   Get the socket type corresponding to a protocol name.
   Params:  protocol                Protocol name - "TCP" or "UDP"
   Returns: netSocketTypeStream     Stream (TCP) socket.
            netSocketTypeDatagram   Datagram (UDP) socket.
            (NetSocketTypeEnum) 0   Error.
*/
static NetSocketTypeEnum
PrvNetCalcSocketType (Char* protocol)
{
   // Declare local variables.
   NetSocketTypeEnum socketType = (NetSocketTypeEnum) 0; // Return value.

   // Check if it's "tcp"
   if (0 == StrCaselessCompare (protocol, "tcp"))
   {
   // Protocol is "tcp".

      // Set socket type to stream.
      socketType = netSocketTypeStream;
   }

   // Check if it's "udp"
   if (0 == StrCaselessCompare (protocol, "udp"))
   {
   // Protocol is "udp".

      // Set socket type to datagram.
      socketType = netSocketTypeDatagram;
   }

   // Return value.
   return socketType;
}

/*
   PrvNetLogSocketInfo:
   Draw a socket's information.
   Params:  socket      Socket whose info to draw.
   Returns: (none)
*/
static void
PrvNetLogSocketInfo (NetSocketRef socket)
{
   // Declare local variables.
   Err err = 0;                           // PalmOS functions' error code.
   Int16 resCode;                         // PalmOS functions' return code.
   NetSocketAddrType locAddr;             // Local address buffer.
   NetSocketAddrType remAddr;             // Remote address buffer.
   NetSocketAddrINType* inetAddrP;        // inet address struct.
   Int16 locAddrLen = sizeof (locAddr);   // Size of locAddr.
   Int16 remAddrLen = sizeof (remAddr);   // Size of remAddr.
   Char addrText [kMaxDotDec+1];     // Address output buffer.

   // Get the address information.
   resCode = NetLibSocketAddr (sNetLibRefNum,
                               socket,
                               &locAddr,
                               &locAddrLen,
                               &remAddr,
                               &remAddrLen,
                               prvNetTimeout,
                               &err);

   // Check for success.
   if (0 == resCode)
   {
   // Success!

      // Draw the subheading.
      PnpLogAddItem (pnpLogNet, "Host Addresses:");

   // Draw local address information.

      // Cast local address to inet address type.
      inetAddrP = (NetSocketAddrINType*) (&locAddr);

      // Convert remote 32-bit host address to dotted decimal.
      NetLibAddrINToA (sNetLibRefNum,
                       inetAddrP->addr,
                       addrText);

      // Format local info string.
      StrPrintF (sText,
                  "\tLocal - %s:%u",
                  addrText,
                  NetNToHS (inetAddrP->port));

      // Draw the local info string.
      PnpLogAddItem (pnpLogNet, sText);

   // Draw remote address information.

      // Cast remote address to inet address type.
      inetAddrP = (NetSocketAddrINType*) (&remAddr);

      // Convert remote 32-bit host address to dotted decimal.
      NetLibAddrINToA (sNetLibRefNum,
                       inetAddrP->addr,
                       addrText);

      // Format remote info string.
      StrPrintF (sText,
                  "\tRemote - %s:%u",
                  addrText,
                  NetNToHS (inetAddrP->port));

      // Draw the remote info string.
      PnpLogAddItem (pnpLogNet, sText);
   }
   else
   {
   // Error getting address info.

      // Draw text error.
      PnpLogAddItem (pnpLogNet, "Error retrieving address info.");
   }
}

/*
   PrvNetSocketDmResize:
   Resize the database record, as necessary, for receipt of data..
   Params:  socketRef   Socket to read data from.
            dataP       Database/record information.
   Returns: >0          Amount of data available for read.
            < 0         Error. No data, record is full or other.
*/
Int16
PrvNetSocketDmResize (NetSocketRef socketRef, GDataPtr dataP)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   Int16 newSize = 0;   // New record size, if necessary.
   Int16 readSize = dataP->recordSize - dataP->recordLength;
   
   // Insist that record be valid in every way.
   ErrFatalDisplayIf (NULL == dataP->recordH, "Record is not available!");
   ErrFatalDisplayIf (false == dataP->canEdit, "Record is read-only!");
   
   // Get current size of record.
   if (readSize >= kMinDataRead)
   {
   // Current record is big enough.
   
      // Return success.
      return readSize;
   }

   // Record does not have enough room left.
   
   // Calculate required new size of record.
   newSize = dataP->recordLength + kMinDataRead;
   
   // Check if new size is possible.
   if (kMaxDataSize < newSize)
   {
   // We cannot grow record large enough.
   
      // Return an error.
      return -1;
   }
   
   // Resize record.
   if (0 != GDtaResizeRecord (dataP, newSize))
   {
   // Resize failed.
   
      // Return an error.
      return -1;
   }
   
   // Resize succeeded. Update size.
   dataP->recordSize = newSize;
   
   // Update readSize, as well.
   readSize = dataP->recordSize - dataP->recordLength;
   
   // Return success.
   return readSize;
}
