/***********************************************************************
 *
 * FileName:
 *		PnpDtp.c
 *
 * Description:
 *    Implementation of DayTime functions for "Palm Network Programming"
 *    sample DayTime Peer application.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Include FTP Function declarations.
#include "PnpDtp.h"
#include "PnpNet.h"
#include "PnpLog.h"
#include "GUnix.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const Char  kCRLF [] = { 0x0D, 0x0A, 0x00 };
static const UInt32 pnpDtpMinInterval = 10;

// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
static Err     PrvDtpSelect (PnpDtpInfoPtr infoP);

static Err     PrvDtpCliOpen (PnpDtpInfoPtr infoP, Char* host);
static Err     PrvDtpCliClose (PnpDtpInfoPtr infoP);
static Err     PrvDtpCliReceive (PnpDtpInfoPtr infoP);
static Err     PrvDtpCliSend (PnpDtpInfoPtr infoP);

static Err     PrvDtpSvrOpen (PnpDtpInfoPtr infoP);
static Err     PrvDtpSvrClose (PnpDtpInfoPtr infoP);
static Err     PrvDtpSvrReceive (PnpDtpInfoPtr infoP);
static Err     PrvDtpSvrSend (PnpDtpInfoPtr infoP);

static Err     PrvDtpGetTimeString (PnpDtpInfoPtr infoP, Char* strP, UInt16 strLen);
static Err     PrvDtpAddLogItem (PnpDtpInfoPtr infoP, Char* itemPrefix, Char* itemText);

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   PnpDtpInit:
   Initialize DayTime peer.
   Params:  infoP    Pointer to DayTime Peer Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpDtpInit (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Initialize Network library.
   err = PnpNetInit ();
   if (err)
   {
      return -1;
   }

   // Initialize FTP session information.
   infoP->isNetConnect = false;
   
   // Initialize daytime client attributes.
   infoP->cliSocket = kNilSocket;
   infoP->cliLastSend = 0;
   
   // Initialize daytime server attributes.
   infoP->svrSocket = kNilSocket;
   MemSet (&infoP->svrPeer, sizeof (infoP->svrPeer), '\0');
   infoP->svrPeerLen = sizeof (infoP->svrPeer);
   infoP->svrIsBusy = false;
   
   // Log successful initialization.
   PrvDtpAddLogItem (infoP, NULL, "INITIALIZE");

   // Return result.
   return 0;
}

/*
   PnpDtpTerm:
   Terminate DayTime peer.
   Params:  infoP    Pointer to DayTime Peer Session information.
   Returns: 0        Success.
            <0       Error code.
*/
Err
PnpDtpTerm (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;      // Return value.

   // Check if we are connected to the network.
   if (false != infoP->isNetConnect)
   {
   // We are connected.
   
      // Check if we can run  (this means we're not running).
      // (We can't be running in if we're not connected)
      if (PnpDtpIsRunning (infoP))
      {
      // We're still logged in.

         // Logout.
         PnpDtpStop (infoP);
      }

      // Disconnect from the network.
      err = PnpNetDisconnect ();

      // Clear network connection flag.
      infoP->isNetConnect = false;
   }
   
   // Terminate network.
   PnpNetTerm ();

   // Log successful initialization.
   PrvDtpAddLogItem (infoP, NULL, "TERMINATE");

   // Return result.
   return err;
}

/*
   PnpDtpProcess:
   Process any incoming or outgoing FTP events.
   Params:  infoP    DayTime Peer Session information.
   Returns: 0        FTP events handled successfully.
            <0       FTP connection ended.
*/
Err
PnpDtpProcess (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;         // Return value.

   // Make sure we are running.
   if (!PnpDtpIsRunning (infoP))
   {
   // Not running.
   
      // Just return an error.
      return -1;
   }
   
   // Check data exchange.
   err = PrvDtpSelect (infoP);
   
   // Check if there is data to read or write.
   if (!err)
   {
   // Sockets are ready.

      // Receive client responses.
      PrvDtpCliReceive (infoP);

      // Send client requests.
      PrvDtpCliSend (infoP);
      
      // Receive server responses.
      PrvDtpSvrReceive (infoP);
      
      // Send server requests.
      PrvDtpSvrSend (infoP);
   }

   // Return error.
   return err;
}

/*
   PnpDtpRun:
   Run the DayTime peer.
   Params:  infoP    DayTime Peer Session object.
            host     Client Peer host name.
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpDtpRun (PnpDtpInfoPtr infoP, Char* host)
{
   // Make sure we are connected to the network.
   if (false == infoP->isNetConnect)
   {
   // We are not connected.
   
      // Connect to the network.
      ErrFatalDisplayIf (0 != PnpNetConnect (),
                      "Error connecting to the network.");

      // Set network connection flag to true.
      infoP->isNetConnect = true;
   }

   // Make sure client opened okay.
   PrvDtpCliOpen (infoP, host);
   
   // Make sure server opened okay.
   PrvDtpSvrOpen (infoP);
   
   // Return value.
   return 0;
}

/*
   PnpDtpStop:
   Logout from FTP Server.
   Params:  infoP    DayTime Peer Session object.
   Returns: 0        Success.
            < 0      Error.
*/
Err
PnpDtpStop (PnpDtpInfoPtr infoP)
{
   // Close client peer.
   PrvDtpCliClose (infoP);
   
   // Close server peer.
   PrvDtpSvrClose (infoP);
   
   // Return success.
   return 0;
}

/*
   PnpDtpIsRunning:
   Check if the we can run the daytime peer.
   Params:  infoP    DayTime Peer Session information.
   Returns: true     DayTime peer is running.
            false    DayTime peer is not running.
*/
Boolean
PnpDtpIsRunning (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Boolean isRunning = false;   // Return value.

   // Control socket must not be connected.
   if (kNilSocket != infoP->cliSocket && kNilSocket != infoP->svrSocket)
   {
   // Client, server sockets are connected.
   
      // We are running.
      isRunning = true;
   }
   
   // Return value.
   return isRunning;   
}

// --------------------------------------------------------------------
// Private function implementations
// --------------------------------------------------------------------

Err
PrvDtpSelect (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   UInt16 width = 0;
   UInt32 now = TimGetSeconds ();
   Err err = -1;
   
   // Initialize the file descriptor sets.
   netFDZero (&infoP->readFDs);
   netFDZero (&infoP->writeFDs);

   // Set Client Socket Descriptors in Read, Write Sets.
   netFDSet (infoP->cliSocket, &infoP->readFDs);
   if (now - infoP->cliLastSend > pnpDtpMinInterval)
   {
   // 10 seconds has passed since last request.
   
      // Check write state of client socket.
      netFDSet (infoP->cliSocket, &infoP->writeFDs);
   }
   
   // Set Client Socket Descriptors in Read, Write Sets.
   if (infoP->svrIsBusy)
   {
   // Server is busy - we have something to send.
   
      // Check write state of server socket.
      netFDSet (infoP->svrSocket, &infoP->writeFDs);
   }
   else
   {
   // Server is not busy - check for requests.
   
      // Check read state of server socket.
      netFDSet (infoP->svrSocket, &infoP->readFDs);
   }
   
   // Calculate width if necessary.
   width = infoP->svrSocket > infoP->cliSocket
           ? infoP->svrSocket
           : infoP->cliSocket; 
   
   // Call PnpNetSelect.
   err = PnpNetSelect (width+1, &infoP->readFDs, &infoP->writeFDs);
   
   // Return results.
   return err;
}


static Err
PrvDtpCliOpen (PnpDtpInfoPtr infoP, Char* host)
{
   // Declare local variables.
   Err err = 0;
   NetSocketAddrType localAddr, remoteAddr;
   NetSocketAddrINType* addrAliasP = NULL;
   
   // Calculate new address.
   err = PnpNetCalcAddress (&remoteAddr, host, "daytime", "udp");
   if (err)
   {
   // Unable to resolve host name.
   
      // Return error.
      return err;
   }
   
   // Open the client socket.
   infoP->cliSocket = PnpNetSocketOpen (netSocketTypeDatagram);
   
   // Insist on success.
   ErrFatalDisplayIf (kNilSocket == infoP->cliSocket,
                      "Error opening DayTime client socket!");

   addrAliasP = (NetSocketAddrINType*) &localAddr;
   
   // Populate address structure.
   addrAliasP->family = netSocketAddrINET;
   addrAliasP->port = NetHToNS (0);
   addrAliasP->addr = NetHToNL (netIPAddrLocal);

   // "Bind" socket to local port.
   err = PnpNetSocketBind (infoP->cliSocket,
                           &localAddr, 
                           sizeof (localAddr));
   
   // Check if socket was bound.
   if (0 == err)
   {
   // Success.

      // Now that we got the socket open, let's set up the client session.
      // We want to "connect" to the server.
      err = PnpNetSocketConnect (infoP->cliSocket, &remoteAddr);
   }
   
   // We're done, return the result.
   return err;
}

static Err
PrvDtpCliClose (PnpDtpInfoPtr infoP)
{
   // Close client socket.
   PnpNetSocketClose (infoP->cliSocket);
   infoP->cliSocket = kNilSocket;
   infoP->cliLastSend = 0;

   // Return success.
   return 0;
}

static Err
PrvDtpCliReceive (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1;
   Char dayTimeString [pnpMaxDayTimeString+1];
   UInt16 dayTimeLen = sizeof (dayTimeString);
   Char* logEntryP = NULL;
   
   // Check if read fd is set for the client socket.
   if (netFDIsSet (infoP->cliSocket, &infoP->readFDs))
   {
   // Read fd is set for client socket.
   
      // Read the data.
      err = PnpNetSocketReceive (infoP->cliSocket,
                         (UInt8*)dayTimeString, 
                                &dayTimeLen);

      // Check for success.
      if (0 == err)
      {
      // Success.

         // Add entry to log.
         PrvDtpAddLogItem (infoP, "Cli Read", dayTimeString);
      }
   }
   
   // Return result.
   return err;
}

static Err
PrvDtpCliSend (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1;
   Char cmdString [] = "GET";
   UInt16 cmdLen = StrLen (cmdString);
   
   // Check if read fd is set for the client socket.
   if (!netFDIsSet (infoP->cliSocket, &infoP->writeFDs))
   {
   // Write fd is not set for client socket.
   
      // We can't send anything so return an error.
      return -1;
   }
   
   // Send the packet.
   err = PnpNetSocketSend (infoP->cliSocket, 
                   (UInt8*)cmdString, 
                          &cmdLen);
   if (!err)
   {
   // Send succeeded.
   
      // Update last send time.
      infoP->cliLastSend = TimGetSeconds ();

      // Add entry to log.
      PrvDtpAddLogItem (infoP, "Cli Send", cmdString);
   }
   
   // Return result.
   return err;
}

static Err
PrvDtpSvrOpen (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   NetSocketAddrType addr;
   
   // Now that we got the socket open, let's set up the server session.
   // We want to "bind" the server to an address.
   err = PnpNetCalcAddress (&addr, "localhost", "daytime", "udp");
   
   // Check if address was calc'd.
   if (err)
   {
   // Unable to resolve address.
   
      // Return error.
      return err;
   }
   
   // Open the server socket.
   infoP->svrSocket = PnpNetSocketOpen (netSocketTypeDatagram);
   
   // Insist on success.
   ErrFatalDisplayIf (kNilSocket == infoP->svrSocket,
                      "Error opening DayTime client socket!");

   // Bind to server address.
   err = PnpNetSocketBind (infoP->svrSocket, &addr, sizeof (addr));
   
   // Return result.
   return err;
}

static Err
PrvDtpSvrClose (PnpDtpInfoPtr infoP)
{   
   // Close server socket.
   PnpNetSocketClose (infoP->svrSocket);
   infoP->svrSocket = kNilSocket;

   // Clear peer attribute.
   MemSet (&infoP->svrPeer, sizeof (infoP->svrPeer), '\0');
   infoP->svrPeerLen = sizeof (infoP->svrPeer);
   
   // Clear the busy flag.
   infoP->svrIsBusy = false;
      
   // Return success.
   return 0;
}   

static Err
PrvDtpSvrReceive (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = -1;
   Char* logEntryP = NULL;
   Char cmdString [pnpMaxCommandString+1];
   UInt16 cmdLen = sizeof (cmdString);
   
   // Check if server is already servicing a request,
   //    or if read fd is not set for the server socket.
   if (infoP->svrIsBusy || !netFDIsSet (infoP->svrSocket, &infoP->readFDs))
   {
   // Either the server is busy, or no data waits to be received.
   
      // Return an error.
      return -1;
   }
   
   // Read the data.
   err = PnpNetSocketReceiveFrom (infoP->svrSocket,
                                 &infoP->svrPeer,
                                 &infoP->svrPeerLen,
                         (UInt8*) cmdString, 
                                 &cmdLen);

   // Check for success.
   if (0 == err)
   {
   // Success.

      // Add entry to log.
      PrvDtpAddLogItem (infoP, "Svr Read", cmdString);
      
      // Set server busy flag.
      infoP->svrIsBusy = true;
   }
   
   // Return result.
   return err;
}

static Err
PrvDtpSvrSend (PnpDtpInfoPtr infoP)
{
   // Declare local variables.
   Err err = 0;
   Char dayTimeString [pnpMaxDayTimeString+1];
   UInt16 dayTimeLen = sizeof (dayTimeString);
   Char* logEntryP = NULL;
   
   // First, we need to check if there is someone to send to.
   if (!infoP->svrIsBusy)
   {
   // Server is not busy -> nothing to send
   
      // Return error code.
      return -1;
   }
   
   // Format time as string.
   PrvDtpGetTimeString (infoP, dayTimeString, dayTimeLen);
   
   // Send time string to peer.
   err = PnpNetSocketSendTo (infoP->svrSocket,
                            &infoP->svrPeer,
                             infoP->svrPeerLen,
                    (UInt8*) dayTimeString, 
                            &dayTimeLen);
   
   if (!err)
   {
   // Send succeeded.
      
      // Add entry to log.
      PrvDtpAddLogItem (infoP, "Svr Send", dayTimeString);
   
      // Server is no longer busy.
      infoP->svrIsBusy = false;
   }
   
   // Return result.
   return err;
}

static Err
PrvDtpGetTimeString (PnpDtpInfoPtr infoP, Char* strP, UInt16 strLen)
{
   Err err = GUnx_ctime (strP, strLen);
   
   if (!err)
   {
      StrCat (strP, kCRLF);
   }
   
   return err;
}

static Err
PrvDtpAddLogItem (PnpDtpInfoPtr infoP, Char* itemPrefix, Char* itemText)
{
   // Declare local variables.
   Err err = 0;      // Return value.
   UInt16 itemEntryLen = 0;
   Char* itemEntryP = NULL;

   if (NULL != itemPrefix)
   {
      itemEntryLen = StrLen (itemPrefix) + StrLen (itemText) + 2 + 1;
      
      // Allocate log entry string.
      itemEntryP = (Char*) MemPtrNew (itemEntryLen);
      
      // Insist on success!
      ErrFatalDisplayIf (NULL == itemEntryP, "Memory Allocation failed!");
      
      // Format entry.
      StrPrintF (itemEntryP, "%s: %s", itemPrefix, itemText);
   }
   else
   {
      itemEntryLen = StrLen (itemText) + 1;

      // Allocate log entry string.
      itemEntryP = (Char*) MemPtrNew (itemEntryLen);
      
      // Insist on success!
      ErrFatalDisplayIf (NULL == itemEntryP, "Memory Allocation failed!");
      
      // Format entry.
      StrPrintF (itemEntryP, "%s", itemText);
   }
   
   err = PnpLogAddItem (pnpLogDtp, itemEntryP);
            
   // Free log entry string.
   MemPtrFree (itemEntryP);   // Add result to log.

   // Return error.
   return err;
}
