/***********************************************************************
 *
 * FileName:
 *		GForm.h
 *
 * Description:
 *    Declaration of Greg's Form API.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

#ifndef __GFORM_H__
#define __GFORM_H__

// Standard standard definitions
#include "GDefs.h"

typedef Boolean (*GFrmEventHandler) (EventPtr eventP);

/*
   GFrmClearEventHandlers
   Clear all event handlers.
   Params:  formId      ID of form whose handlers to set. Currently ignored.
   Returns: 0           Success
            ! 0         Error
*/
Err GFrmClearEventHandlers (UInt16 formId);

/*
   GFrmSetEventHandler:
   Set the handler for a form event.
   Params:  formId      ID of form whose handlers to set. Currently ignored.
            eventID     ID of event with which to associate handler.
            handlerP    Event handler function.
   Returns: 0           Success.
            ! 0         Error.
*/
Err GFrmSetEventHandler (UInt16 formId, UInt16 eventID, GFrmEventHandler handlerP);

/*
   GFrmHandleEvent:
   Handle a form event.
   Params:  eventP      Event to handle.
   Returns: true        Event handled successfully and completely.
            false       Event not handled completely and successfully.
   Note:    Applications should set form event handler to this
            function when loading new forms.
*/
Boolean GFrmHandleEvent (EventPtr eventP);

/*
   GFrmGetControl:
   Get the pointer to a control object.
   Params:  formP    Form containing control.
            controlID  ID of control to retrieve.
   Returns: ! NULL   Pointer to control object.
            NULL     An error occurred.
*/
ControlPtr GFrmGetControl (FormPtr formP, UInt16 controlID);

/*
   GFrmShowControl:
   Show or hide a control object.
   Params:  formP       Form containing control.
            controlID   ID of control control to show or hide.
            showFlag    true to show control, false to hide.
   Returns: 0           Pointer to control object.
            ! 0         An error occurred.
*/
Err GFrmShowControl (FormPtr formP, UInt16 controlID, Boolean showFlag);

/*
   GFrmGetField:
   Get the pointer to a field object.
   Params:  formP    Form containing field.
            fieldID  ID of field control to retrieve.
   Returns: ! NULL   Pointer to field object.
            NULL     An error occurred.
*/
FieldPtr GFrmGetField (FormPtr formP, UInt16 fieldID);

/*
   GFrmGetFieldTextHandle:
   Get a handle to a form field's text memory.
   Params:  formP    Form from which to retrieve text.
            fieldID  ID of field from which to retrieve text.
   Returns: ! 0      Handle to text data block.
            ! 0      Error.
*/
MemHandle GFrmGetFieldTextHandle (FormPtr formP, UInt16 fieldID);

/*
   GFrmGetFieldText:
   Get the text (or text length) from a form field.
   Params:  formP    Form from which to retrieve text.
            fieldID  ID of field from which to retrieve text.
            textP    Buffer for field text. NULL to retrieve size.
            sizeP    [in]  size of text buffer. 0 to retrieve size.
                     [out] number of bytes copied, including nil
                           terminator. Or field text's size if textP
                           is NULL.
   Returns: 0        Success.
            ! 0      Error.
*/
Err GFrmGetFieldText (FormPtr formP, UInt16 fieldID, Char* textP, UInt16* sizeP);

/*
   GFrmSetFieldText:
   Set the text of a form field.
   Params:  formP    Form containing field to set.
            fieldID  ID of field to set.
            textP    New text for the field.
   Returns: 0        Success.
            ! 0      Error.
*/
Err GFrmSetFieldText (FormPtr formP, UInt16 fieldID, Char* textP);

/*
   GFrmGetList:
   Get the pointer to a list object.
   Params:  formP    Form containing list.
            listID  ID of list control to retrieve.
   Returns: ! NULL   Pointer to list object.
            NULL     An error occurred.
*/
ListPtr GFrmGetList (FormPtr formP, UInt16 listID);

/*
   GFrmGetListText:
   Get the text (or text length) from a form list.
   Params:  formP    Form from which to retrieve text.
            listID  ID of list from which to retrieve text.
            textP    Buffer for list text. NULL to retrieve size.
            sizeP    [in]  size of text buffer. 0 to retrieve size.
                     [out] number of bytes copied, including nil
                           terminator. Or list text's size if textP
                           is NULL.
   Returns: 0        Success.
            ! 0      Error.
*/
Err GFrmGetListText (FormPtr formP, UInt16 listID, Char* textP, UInt16* sizeP);

/*
   GFrmGetTable:
   Get the pointer to a table object.
   Params:  formP    Form containing table.
            tableID  ID of table control to retrieve.
   Returns: ! NULL   Pointer to table object.
            NULL     An error occurred.
*/
TablePtr GFrmGetTable (FormPtr formP, UInt16 tableID);

/*
   GFrmRedrawTable:
   Redraw a table on a form.
   Params:  formID   Form containing table.
            tableID  ID of table control to redraw.
   Returns: 0        Success.
            ! 0      Error.
   
*/
Err GFrmRedrawTable (FormPtr formP, UInt16 tableID);

#endif	// __GFORM_H__
