/***********************************************************************
 *
 * FileName:
 *		GData.c
 *
 * Description:
 *    Implementation of GLib database record support.
 *
 * Copyright  2001, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Standard application definitions
#include "GData.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
static const UInt16 kCardNo = 0;
static const UInt32 kDBType = 'data';

// --------------------------------------------------------------------
// Private typedef declarations
// --------------------------------------------------------------------

// --------------------------------------------------------------------
// Private structure declarations
// --------------------------------------------------------------------
/*
struct GDataType
{
   UInt16 recordIndex;
   MemHandle recordH;
   UInt16 recordSize;
   UInt16 recordLength;
   Boolean canEdit;
};
*/

// --------------------------------------------------------------------
// Private variables
// --------------------------------------------------------------------
DmOpenRef s_dbR;

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   GDtaCreate:
   Open a Palm database. If the database does not exist,
   create it.
   Params:  dataP       Database/record information.
            creatorId   Creator id of database to open.
            name        Name of database to open.
   Returns: 0           Success.
            <0          Error code.
*/
Err
GDtaOpen (GDataPtr dataP, UInt32 creatorID, Char* name)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   LocalID localID = 0;
   
   ErrFatalDisplayIf (s_dbR, "Database already open!");

   // Initialize out-param.
   MemSet (dataP, sizeof (GDataType), '\0');
   dataP->recordIndex = glibInvalidIndex;
   
   // Look for database.
   localID = DmFindDatabase (kCardNo, name);

   // Check if database was found.
   if (0 == localID)
   {
   // Database not found.

      // Create log database.
      err = DmCreateDatabase (kCardNo, name, creatorID, kDBType, false);
   }

   // Check for error.
   if (!err)
   {
   // Success to here.

      // Try to open database.
      s_dbR = DmOpenDatabaseByTypeCreator (kDBType,
                                           creatorID,
                                           dmModeReadWrite);

      // Check if opened successfully.
      if (!s_dbR)
      {
         // Set error code.
         err = -1;
      }
   }

   // Return result.
   return err;
}

/*
   GDtaClose:
   Close the Database.
   Params:  dataP       Database/record information.
   Returns: 0           Success.
            <0          Error code.
*/
Err GDtaClose (GDataPtr dataP)
{
   // Declare local variables.
   Err err = 0;
   
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   // Release the current record, if any.
   GDtaReleaseRecord (dataP);
   
   // Close the database.
   DmCloseDatabase (s_dbR); s_dbR = NULL;

   // Clear out-param.
   MemSet (dataP, sizeof (GDataType), '\0');
   dataP->recordIndex = glibInvalidIndex;

   return err;
}
   
/*
   GDtaEmpty:
   Remove all records from a Database.
   Params:  dataP       Database/record information.
   Returns: 0           Success.
            <0          Error code.
*/
Err
GDtaEmpty (GDataPtr dataP)
{
   // Declare local variables.
   Err err = 0;
   UInt16 category;
   
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   for (category = 0; category < 16; category++)
   {
      DmDeleteCategory (s_dbR, category);
   }
   
   dataP->recordIndex = glibInvalidIndex;
   dataP->recordSize = 0;
   dataP->recordLength = 0;

   return err;
}

/*
   GDtaNumRecords:
   Get the number of records in the database.
   Params:  dataP       Database/record information.
   Returns: >=0         Number of entries on the queue.
*/
UInt16
GDtaNumRecords (GDataPtr /*dataP*/)
{
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   // Return the number of records.
   return DmNumRecords (s_dbR);
}

/*
   GDtaAppendRecord:
   Append a record to database.
   Params:  dataP       Database/record information.
            headerSize  Size of record header, 0 if none.
   Returns: 0           Success.
            <0          Error
*/
Err
GDtaAppendRecord (GDataPtr dataP, UInt16 recordSize)
{
   // Declare local values.
   MemPtr recordP = NULL;
   
   // Release old record, if any.
   GDtaReleaseRecord (dataP);
   
   // Get number of records.
   dataP->recordIndex = GDtaNumRecords (dataP);
   
   // Create a new record.
   dataP->recordH = DmNewRecord (s_dbR, &dataP->recordIndex, recordSize);
   
   // Check for success.
   if (NULL == dataP->recordH)
   {
   // Failure.
   
      // Return error.
      return -1;
   }

   // Update record information.
   dataP->recordSize = recordSize;
   dataP->recordLength = 0;

   // Record can be edited - we have it's busy bit.
   dataP->canEdit = true;
   
   // Lock the record.
   recordP = MemHandleLock (dataP->recordH);
   
   DmSet (recordP, 0, recordSize,'\0');

   // Unlock the record.
   MemPtrUnlock (recordP);
   
   // Return success.
   return 0;
}

/*
   GDtaResizeRecord:
   Resize a record in the database.
   Params:  dataP       Database/record information.
            newSize     Requested new size.
   Returns: 0           Success
            <0          Error
*/
Err
GDtaResizeRecord (GDataPtr dataP, UInt16 newSize)
{
   // Declare local variables.
   Err err = 0;
   
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   // Insist that we have valid database/record info.
   ErrFatalDisplayIf (!dataP, "Database/Record Information is not valid!");
   
   // Insist that record index is valid.
   ErrFatalDisplayIf (glibInvalidIndex == dataP->recordIndex,
                     "Record index is invalid!");
   
   // Check if record is editable.
   if (false == dataP->canEdit)
   {
   // Can't edit.
   
      // Return error.
      return -1;
   }
   
   // Check if we are already the right size.
   if (dataP->recordSize == newSize)
   {
   // Already the right size.
   
      // Return success.
      return 0;
   }
                     
   // Okay, now we can get to work!
   dataP->recordH = DmResizeRecord (s_dbR, dataP->recordIndex, newSize);
   
   // Check if resize failed.
   if (NULL == dataP->recordH)
   {
   // Resize failed.
   
      // Return error.
      return -1;
   }
   
   // Resize succeeded.
   
   // Update database/record information.
   dataP->recordSize = newSize;
   dataP->canEdit = true;
   
   // Return success.
   return 0;
}

/*
   GDtaQueryRecord:
   Retrieve the record from the database. The record is
   read only.
   Params:  dataP       Database/record information.
            recordIndex Index of record to retrieve.
   Returns: 0           Success
            <0          Error
*/
Int16 GDtaQueryRecord (GDataPtr dataP, UInt16 recordIndex)
{
   // Declare local variables.
   Err err = 0;
   
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   // Insist that we have valid database/record info.
   ErrFatalDisplayIf (!dataP, "Database/Record Information is not valid!");
   
   // Insist that record index is valid.
   ErrFatalDisplayIf (glibInvalidIndex == recordIndex,
                     "Record index is invalid!");
                     
   // Okay, now we can get to work!

   // Check if we are already pointing to the record.
   if (dataP->recordIndex == recordIndex && NULL != dataP->recordH)
   {
   // Already got it.
   
      // Return success.
      return 0;
   }
   
   // Need to get it.   
   dataP->recordH = DmQueryRecord (s_dbR, recordIndex);
   
   // Check if query failed.
   if (NULL == dataP->recordH)
   {
   // Query failed.
   
      // Return error.
      return -1;
   }
   
   // Query succeeded.
   
   // Update database/record information.
   dataP->recordLength = dataP->recordSize = MemHandleSize (dataP->recordH);
   dataP->canEdit = false;
   
   // Return success.
   return 0;
}

/*
   GDtaReleaseRecord:
   Release a record.
   Params:  dataP       Database/record information.
   Returns: 0           Success.
            <0          Error.
*/
Err
GDtaReleaseRecord (GDataPtr dataP)
{
   // Declare local variables.
   Err err = 0;
   
   // Insist that database is open.
   ErrFatalDisplayIf (!s_dbR, "Database is not open!");
   
   // Insist that we have valid database/record info.
   ErrFatalDisplayIf (!dataP, "Database/Record Information is not valid!");
                     
   // Okay, now we can get to work!
   
   // Make sure we have a record.
   if (glibInvalidIndex == dataP->recordIndex || NULL == dataP->recordH)
   {
   // We have no record.
   
      // Return error.
      return -1;
   }
   
   // Check if we were in edit mode.
   if (false == dataP->canEdit)
   {
   // Can't edit.
   
      // Clear record handle attribute.
      dataP->recordH = NULL;
      
      // Return success.
      return 0;
   }
   
   // First, make sure length and size are the same.
   GDtaResizeRecord (dataP, dataP->recordLength);
   
   // Release the record.   
   DmReleaseRecord (s_dbR, dataP->recordIndex, dataP->canEdit);
   dataP->recordH = NULL;
   
   // Update database/record information.
   dataP->recordIndex = glibInvalidIndex;
   dataP->recordLength = dataP->recordSize = 0;
   dataP->canEdit = false;
   
   // Return success.
   return 0;
}

