/***********************************************************************
 *
 * FileName:
 *		GForm.c
 *
 * Description:
 *    Implementation of GLib Form support.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Form function declarations.
#include "GForm.h"

// Private type declarations.
typedef struct PrvEventMapping
{
   UInt16 eventID;
   GFrmEventHandler handlerP;
} PrvEventMapping;

// Private function declarations.
static void* PrvGetObjectPtr (FormPtr formP, UInt16 controlID);
static PrvEventMapping* PrvGetEventMapping (UInt16 eventID);

// Private constant declarations.
const kMaxEventMappings = 20;

// Private static data.
static PrvEventMapping eventMappings [kMaxEventMappings];
static UInt16 numEventMappings = 0;

/*
   GFrmClearEventHandlers
   Clear all event handlers.
   Params:  formId      ID of form whose handlers to set. Currently ignored.
   Returns: 0           Success
            ! 0         Error
*/
Err
GFrmClearEventHandlers (UInt16 formId)
{
   // Clear all handlers.
   MemSet (eventMappings,
           kMaxEventMappings * sizeof (PrvEventMapping),
           '\0');
           
   // Set event handler count to 0.
   numEventMappings = 0;
   
   // Return success.
   return 0;
}

/*
   GFrmSetEventHandler:
   Set the handler for a form event.
   Params:  formId      ID of form whose handlers to set. Currently ignored.
            eventID     ID of event with which to associate handler.
            handlerP    Event handler function.
   Returns: 0           Success.
            ! 0         Error.
*/
Err
GFrmSetEventHandler (UInt16 formId, UInt16 eventID, GFrmEventHandler handlerP)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   PrvEventMapping* pMap = NULL;
   
   // Check if there is already a handler for this event.
   if (NULL != (pMap = PrvGetEventMapping (eventID)))
   {
   // There is an event handler. 
   
      // Set new handler.
      pMap->handlerP = handlerP;
   }
   
   // No handler set, check if we can add one.
   else if (numEventMappings < kMaxEventMappings)
   {
   // We still have room.
   
      // Add new handler.
      pMap = &eventMappings [numEventMappings++];
      
      // Set event handler attributes.
      pMap->eventID = eventID;
      pMap->handlerP = handlerP;
   }
   
   // No existing handler and no room for another.
   else
   {
   // Set error.
   
      err = glibErrFull;
   }
   
   // Return value.
   return err;
}

/*
   GFrmHandleEvent:
   Handle a form event.
   Params:  eventP      Event to handle.
   Returns: true        Event handled successfully and completely.
            false       Event not handled completely and successfully.
   Note:    Applications should set form event handler to this
            function when loading new forms.
*/
Boolean
GFrmHandleEvent (EventPtr eventP)
{
   // Declare local variables.
   Boolean handled = false;      // Return value.
   PrvEventMapping* pMap = NULL;
   
   // Check if there is already a handler for this event.
   if (NULL != (pMap = PrvGetEventMapping (eventP->eType)))
   {
   // There is an event handler. 
   
      // Invoke event handler to handle event.
      handled = pMap->handlerP (eventP);
   }
   
   // Return value.
   return handled;
}

/*
   GFrmGetControl:
   Get the pointer to a control object.
   Params:  formP       Form containing control.
            controlID   ID of control control to retrieve.
   Returns: ! NULL      Pointer to control object.
            NULL        An error occurred.
*/
ControlPtr
GFrmGetControl (FormPtr formP, UInt16 controlID)
{
   // Get the control pointer.
   ControlPtr controlP = (ControlPtr) PrvGetObjectPtr (formP, controlID);

   // Return the control pointer.
   return controlP;
}

/*
   GFrmShowControl:
   Show or hide a control object.
   Params:  formP       Form containing control.
            controlID   ID of control control to show or hide.
            showFlag    true to show control, false to hide.
   Returns: 0           Pointer to control object.
            ! 0         An error occurred.
*/
Err
GFrmShowControl (FormPtr formP, UInt16 controlID, Boolean showFlag)
{
   // Declare local variables.
   Err err = 0;         // Return value.
   ControlPtr controlP = NULL;
	
   // Get the control pointer.
   controlP = GFrmGetControl (formP, controlID);

   // Show or hide the control.
   if (showFlag)
   {
   // Show the control.
      CtlShowControl (controlP);
   }
   else
   {
   // Hide the control.
      CtlHideControl (controlP);
   }
   
   // Return value.
   return err;
}

/*
   GFrmGetField:
   Get the pointer to a field object.
   Params:  formP    Form containing field.
            fieldID  ID of field control to retrieve.
   Returns: ! NULL   Pointer to field object.
            NULL     An error occurred.
*/
FieldPtr
GFrmGetField (FormPtr formP, UInt16 fieldID)
{
   // Declare local variables.
   FieldPtr fieldP = (FieldPtr) PrvGetObjectPtr (formP, fieldID);

   // Return the field pointer.
   return fieldP;
}

/*
   GFrmGetFieldTextHandle:
   Get a handle to a form field's text memory.
   Params:  formP    Form from which to retrieve text.
            fieldID  ID of field from which to retrieve text.
   Returns: ! 0      Handle to text data block.
            ! 0      Error.
*/
MemHandle
GFrmGetFieldTextHandle (FormPtr formP, UInt16 fieldID)
{
   // Declare local variables.
   FieldPtr fieldP = NULL;
   MemHandle fieldTextH = NULL;

   // Get the field object.
   fieldP = GFrmGetField (formP, fieldID);

   // Retrieve the text handle for the field.
   fieldTextH = FldGetTextHandle (fieldP);
   
   // Return value.
   
   return fieldTextH;
}

/*
   GFrmGetFieldText:
   Get the text (or text length) from a form field.
   Params:  formP    Form from which to retrieve text.
            fieldID  ID of field from which to retrieve text.
            textP    Buffer for field text. NULL to retrieve textSize.
            sizeP    [in]  textSize of text buffer. 0 to retrieve textSize.
                     [out] number of bytes copied, including nil
                           terminator. Or field text's textSize if textP
                           is NULL.
   Returns: 0        Success.
            ! 0      Error.
*/
Err
GFrmGetFieldText (FormPtr formP, UInt16 fieldID, Char* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   MemHandle fieldTextH = NULL;
   Char* fieldTextP = NULL;
   UInt16 textSize = 0;

   // Retrieve the text handle for the field.
   fieldTextH = GFrmGetFieldTextHandle (formP, fieldID);
   
   // Check if text handle is valid.
   if (NULL == fieldTextH)
   {
   // Text handle is not valid.
   
      // Check if caller wants text or size.
      if (NULL != textP)
      {
      // Caller size and text.
      
         // Clear text buffer.
         *textP = '\0';
      }
      
      // Set size out-param.
      *sizeP = 0;
   
      // Return.
      return 0;
   }
   
   // Lock the text handle.
   fieldTextP = (Char*) MemHandleLock (fieldTextH);
   
   // Get the textSize of the text.
   textSize = StrLen (fieldTextP) + 1; // Remember Nil-terminator.
   
   // Check if caller wants value or size.
   if (NULL == textP)
   {
   // Caller wants the textSize.
   
      // Set out-value to required size.
      *sizeP = textSize;
   }
   else
   {
   // Caller wants text.
   
      // Make sure buffer is big enough.
      if (*sizeP < textSize)
      {
      // Buffer is too small.
      
         // Set error (but still populate buffer).
         err = glibErrFull;
         
         // Update local textSize value.
         textSize = *sizeP;
      }
      else
      {
      // Buffer is big enough
      
         // Update textSize out-value.
         *sizeP = textSize;
      }
      
      // Note: At this point, *sizeP == textSize.
      
      // Copy field text to out buffer.
      StrNCopy (textP, fieldTextP, textSize - 1);
      
      // Terminate the string
      textP [textSize-1] = '\0';
   }

   // Return value.
   return err;   
}

/*
   GFrmSetFieldText:
   Set the text of a form field.
   Params:  formP    Form containing field to set.
            fieldID  ID of field to set.
            textP    New text for the field.
   Returns: 0        Success.
            ! 0      Error.
*/
Err
GFrmSetFieldText (FormPtr formP, UInt16 fieldID, Char* textP)
{
   // Declare local variables.
   Err err = 0;               // Return value.
   FieldPtr fieldP = NULL;
   MemHandle newTextH = NULL;
   Char* newTextP;
   UInt16 textSize;

   // Calc. length of new text.
   textSize = StrLen (textP) + 1; // Don't forget nil-terminator.

   // Get the field object.
   fieldP = GFrmGetField (formP, fieldID);

   // Retrieve the text handle for the field.
   newTextH = FldGetTextHandle (fieldP);

   // Check if the text handle is valid.
   if (NULL != newTextH)
   {
   // Already a buffer allocated.

      // Set the field's text handle to NULL (just to be safe).
      FldSetTextHandle (fieldP, NULL);

      // Free the old buffer.
      MemHandleFree (newTextH);
   }

   // Allocate new buffer.
   newTextH = MemHandleNew (textSize);

   // Make sure alloc worked.
   ErrFatalDisplayIf (NULL == newTextH, "Memory allocation failed.");

   // Lock new buffer.
   newTextP = (Char*) MemHandleLock (newTextH);
   
   // Copy new value to buffer.
   StrCopy (newTextP, textP);

   // Unlock the buffer.
   MemHandleUnlock (newTextH);

   // Set new buffer in field.
   FldSetTextHandle (fieldP, newTextH);

   // Return value.
   return err;
}

/*
   GFrmGetList:
   Get the pointer to a field object.
   Params:  formP    Form containing field.
            fieldID  ID of field control to retrieve.
   Returns: ! NULL   Pointer to field object.
            NULL     An error occurred.
*/
ListPtr
GFrmGetList (FormPtr formP, UInt16 fieldID)
{
   // Declare local variables.
   ListPtr fieldP = (ListPtr) PrvGetObjectPtr (formP, fieldID);

   // Return the field pointer.
   return fieldP;
}

/*
   GFrmGetListText:
   Get the selected text (or text length) from a form list.
   Params:  formP    Form from which to retrieve text.
            listID  ID of field from which to retrieve text.
            textP    Buffer for field text. NULL to retrieve textSize.
            sizeP    [in]  textSize of text buffer. 0 to retrieve textSize.
                     [out] number of bytes copied, including nil
                           terminator. Or field text's textSize if textP
                           is NULL.
   Returns: 0        Success.
            ! 0      Error.
*/
Err
GFrmGetListText (FormPtr formP, UInt16 listID, Char* textP, UInt16* sizeP)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   ListPtr listP = NULL;
   UInt16 selectedItem = 0;
   Char* selectionP = NULL;
   UInt16 textSize = 0;

   // Get the list object.
   listP = GFrmGetList (formP, listID);

   // Get the selected item.
   selectedItem = LstGetSelection (listP);

   // Make sure there was a selection.
   if (noListSelection != selectedItem)
   {
      // Retrieve selected item's text.
      selectionP = LstGetSelectionText (listP, selectedItem);
   }

   // Get the textSize of the text.
   textSize = StrLen (selectionP) + 1; // Remember Nil-terminator.
   
   // Check if caller wants value or textSize.
   if (NULL == textP || 0 == *sizeP)
   {
   // Caller wants the textSize.
   
      // Set out-value to required textSize.
      *sizeP = textSize;
   }
   else
   {
   // Caller wants text.
   
      // Make sure buffer is big enough.
      if (*sizeP < textSize)
      {
      // Buffer is too small.
      
         // Set error (but still populate buffer).
         err = glibErrFull;
         
         // Update local textSize value.
         textSize = *sizeP;
      }
      else
      {
      // Buffer is big enough
      
         // Update textSize out-value.
         *sizeP = textSize;
      }
      
      // Note: At this point, *sizeP == textSize.
      
      // Copy field text to out buffer.
      StrNCopy (textP, selectionP, textSize - 1);
      
      // Terminate the string
      textP [textSize-1] = '\0';
   }

   // Return value.
   return err;   
}

/*
   GFrmGetTable:
   Get the pointer to a table object.
   Params:  formP    Form containing table.
            tableID  ID of table control to retrieve.
   Returns: ! NULL   Pointer to table object.
            NULL     An error occurred.
*/
TablePtr
GFrmGetTable (FormPtr formP, UInt16 tableID)
{
   // Declare local variables.
   TablePtr tableP = (TablePtr) PrvGetObjectPtr (formP, tableID);

   // Return the table pointer.
   return tableP;
}

/*
   GFrmRedrawTable:
   Redraw a table on a form.
   Params:  formID   Form containing table.
            tableID  ID of table control to redraw.
   Returns: 0        Success.
            ! 0      Error.
   
*/
Err
GFrmRedrawTable (FormPtr formP, UInt16 tableID)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   
   // Get a pointer to the table object.
   TablePtr tableP = GFrmGetTable (formP, tableID);

   // Erase the table.
	TblEraseTable (tableP);

   // Redraw the table.
	TblDrawTable (tableP);
	
	// Return value.
	return err;
}

static void*
PrvGetObjectPtr (FormPtr formP, UInt16 id)
{
   UInt16 index = FrmGetObjectIndex (formP, id);
   void* p = FrmGetObjectPtr (formP, index);
   return p;
}

static PrvEventMapping*
PrvGetEventMapping (UInt16 eventID)
{
   // Declare local variables.
   PrvEventMapping* pMap = NULL;    // Return value.
   UInt16 i;
   
   for (i = 0; i < numEventMappings; i++)
   {
      if (eventID == eventMappings [i].eventID)
      {
      // We found the event id.
      
         // Assign handler to return value.
         pMap = eventMappings + i;
      } 
   }
   
   // Return value.
   return pMap;
}
