/***********************************************************************
 *
 * FileName:
 *		GQueue.c
 *
 * Description:
 *    Implementation of GLib Queue support.
 *
 * Copyright  2000, Greg Winton.
 * All rights reserved.
 *
 *******************************************************************/

// Queue function declarations.
#include "GQueue.h"

// --------------------------------------------------------------------
// Private constant declarations
// --------------------------------------------------------------------
 
// --------------------------------------------------------------------
// Private function declarations
// --------------------------------------------------------------------
Int16* PrvLockQueue (GQueuePtr qP);
void PrvUnlockQueue (GQueuePtr qP);
MemHandle PrvInvokeClone (GQueuePtr qP, MemPtr entryP);
Err PrvInvokeDelete (GQueuePtr qP, MemHandle entryH);
MemHandle PrvCloneEntry (MemPtr entryP);
Err PrvDeleteEntry (MemHandle entryH);

// --------------------------------------------------------------------
// Private data declarations
// --------------------------------------------------------------------

// --------------------------------------------------------------------
// Public function implementations
// --------------------------------------------------------------------

/*
   GQueInit:
   Initialize a queue
   Params:  qP          Queue instance.
            cloneFunc   Function for cloning entries.
            deleteFunc  Function for deleting entries.
            queueSize   Initial size of queue.
   Returns: 0        Success.
            <0       Error code.
*/
Err
GQueInit (GQueuePtr qP, UInt16 queueSize)
{
   // Declare local variables.
   Err err = 0;         // Return value.

   // Allocate an array of memory handles.
   qP->theQueue = MemHandleNew (queueSize * sizeof (Int16));

   // Check if memory allocated.
   if (qP->theQueue)
   {
   // Queue instance initialized.

      // Initialize other attributes.
      qP->queueSize = queueSize;
      qP->numEntries = 0;
   }
   else
   {
   // Error allocating queue instance.

      // Set return value accordingly.
      err = glibErrMemory;
   }

   // Return value.
   return err;
}

/*
   GQueTerm:
   Terminate a queue instance.
   Params:  qP       Queue instance.
   Returns: 0        Success.
            <0       Error code.
*/
Err
GQueTerm (GQueuePtr qP)
{
   // Check if queue is allocated.
   if (qP->theQueue)
   {
      // Clear the queue.
      GQueClear (qP);

      // Free the queue.
      MemHandleFree (qP->theQueue);

      // Queue is no longer valid.
      qP->theQueue = NULL;
   }
   
   // Terminate rest of attributes.
   qP->queueSize = 0;
   qP->numEntries = 0;

   // Return success.
   return 0;
}

/*
   GQueClear:
   Clear all entries from a queue instance.
   Params:  qP       Queue instance.
   Returns: 0        Success.
            <0       Error code.
*/
Err
GQueClear (GQueuePtr qP)
{
   // Declare local variables.
   Int16* entryArray = NULL;

   // Lock the queue.
   entryArray = PrvLockQueue (qP);

   // Assume lock succeeded.
   ErrFatalDisplayIf (NULL == entryArray, "Error locking memory.");

   MemSet (entryArray, qP->numEntries * sizeof entryArray[0], '\0');

   // Set number of entries to 0.
   qP->numEntries = 0;

   // Unlock the queue.
   PrvUnlockQueue (qP);
   
   // Return success.
   return 0;
}

/*
   GQueNumEntries:
   Get the number of entries on the queue.
   Params:  qP       Queue instance.
   Returns: >=0      Number of entries on the queue.
*/
UInt16
GQueNumEntries (GQueuePtr qP)
{
   // Return number of entries.
   return qP->numEntries;
}

/*
   GQueEnqueue:
   Enqueue an entry on the queue.
   Params:  qP             Pointer to FTP Session information.
            entryP         Entry to enqueue.
   Returns: 0              Success.
            glibErrFull    No room for queue item.
            glibErrMemory  Unable to clone item.
*/
Err
GQueEnqueue (GQueuePtr qP, Int16 entry)
{
   // Declare local variables.
   Err err = 0;            // Return value.
   Int16* entryArray;

   // Check if we have room.
   if (qP->numEntries < qP->queueSize)
   {
   // There's room.

      // Lock down the queue.
      entryArray = PrvLockQueue (qP);

      // Assume lock succeeded.
      ErrFatalDisplayIf (NULL == entryArray, "Error locking memory.");

      // Set value in queue.
      entryArray [qP->numEntries] = entry;

      // Increment number of entries.
      ++qP->numEntries;

      // Unlock the queue.
      PrvUnlockQueue (qP);

      // Assume return value is 0.
      ErrFatalDisplayIf (0 != err, "Invalid return value.");
   }
   else
   {
   // No room for more entries.

      // Someday we may try to grow the array, but for now...

      // Set error to full.
      err = glibErrFull;
   }

   // Return value.
   return err;
}

/*
   GQueDequeue:
   Dequeue an entry on the queue.
   Params:  qP             Pointer to FTP Session information.
            entryP         Entry to enqueue.
   Returns: 0              Success.
            glibErrNoData  No entries on the queue.
   Note:    The caller of this function takes ownership of the returned
            memory handle; it is the callers responsibility to free this
            handle when it is done with it. (Greg, 7/26/00)
*/
Int16
GQueDequeue (GQueuePtr qP)
{
   // Declare local variables.
   Err err = 0;		// Return value.
   Int16 entry, * entryArray;
   UInt16 moveSize = 0;

   // Check if we have any entries.
   if (0 < qP->numEntries)
   {
   // There're entries.

      // Lock down the queue.
      entryArray = PrvLockQueue (qP);

      // Assume lock succeeded.
      ErrFatalDisplayIf ((NULL == entryArray), "Error locking memory.");

      // Set value in queue.
      entry = entryArray [0];

      // Calc. size of move.
      moveSize = (--qP->numEntries) * sizeof (UInt16);
      
      // Move array entries.
      MemMove (entryArray, entryArray + 1, moveSize);  // Decrement entry count

      // Set last entry to NULL.
      entryArray [qP->numEntries] = NULL;

      // Unlock the queue.
      PrvUnlockQueue (qP);
   
      // Assume return value is 0.
      ErrFatalDisplayIf (0 != err, "Invalid return value.");
   }
   else
   {
   // No more entries.

      // Set error.
      err = glibErrNoData;
   }

   // Return value.
   return entry;
}

/*
   GQuePeek:
   Peek at the first entry on the queue.
   Params:  qP             Pointer to FTP Session information.
            entryP         Entry to enqueue.
   Returns: 0              Success.
            glibErrNoData  No entries on the queue.
   Note:    The caller of this function *does not* takes ownership of the
            returned memory handle; it is the callers responsibility to
            free this handle when it is done with it. (Greg, 7/26/00)
*/
Int16
GQuePeek (GQueuePtr qP)
{
   // Declare local variables.
   Err err = 0;
   Int16 entry, * entryArray;

   // Check if we any items.
   if (0 < qP->numEntries)
   {
   // There's room.

      // Lock down the queue.
      entryArray = PrvLockQueue (qP);

      // Assume lock succeeded.
      ErrFatalDisplayIf (entryArray, "Error locking memory.");

      // Set value in queue.
      entry = entryArray [qP->numEntries];

      // Unlock the queue.
      PrvUnlockQueue (qP);
   
      // Assume return value is 0.
      ErrFatalDisplayIf (0 != err, "Invalid return value.");
   }
   else
   {
   // No more entries.

      // Set error.
      err = glibErrNoData;
   }

   // Return value.
   return entry;
}

Int16*
PrvLockQueue (GQueuePtr qP)
{
   // Declare local variables.
   Int16* entryArray;

   // Lock the queue.
   entryArray = (Int16*) MemHandleLock (qP->theQueue);

   // Return value.
   return entryArray;
}
   
void
PrvUnlockQueue (GQueuePtr qP)
{
   // Unlock the queue.
   MemHandleUnlock (qP->theQueue);
}

#ifdef DEFAULT_TO_MEMORY

MemHandle
PrvInvokeClone (GQueuePtr qP, MemPtr entryP)
{
   // Must have a Clone function.
   ErrFatalDisplayIf (!qP->cloneFunc, "Invalid queue object.");

   // Invoke callback.
   return qP->cloneFunc (entryP);
}

Err
PrvInvokeDelete (GQueuePtr qP, MemHandle entryH)
{
   // Must have a Delete function.
   ErrFatalDisplayIf (!qP->deleteFunc, "Invalid queue object.");

   // Invoke callback.
   return qP->deleteFunc (entryH);
}

/*
   PrvCloneEntry:
   Default queue entry clone function.
   Params:  entryP         Entry to clone.
   Returns: (MemHandle)    Clone of entryP or NULL if error.
*/
MemHandle
PrvCloneEntry (MemPtr entryP)
{
   // Declare local variables.
   MemHandle cloneH = NULL;   // Return value.
   MemPtr cloneP = NULL;
   UInt16 size = 0;

   // Determine size of old entry.
   size = MemPtrSize (entryP);

   // Allocate buffer.
   cloneH = MemHandleNew (size);

   // Check if buffer allocated.
   if (NULL != cloneH)
   {
   // Buffer allocated successfully.

      // Lock down clone.
      cloneP = MemHandleLock (cloneH);

      // Copy original to clone.
      MemMove (cloneP, entryP, size);

      // Unlock the clone.
      MemHandleUnlock (cloneH);

      // Clone pointer no longer valid.
      cloneP = NULL;
   }

   // Return value.
   return cloneH;
}

Err
PrvDeleteEntry (MemHandle entryH)
{
   // Just free the memory.
   MemHandleFree (entryH);

   // Return success.
   return 0;
}

#else

/*
   PrvCloneEntry:
   Default queue entry clone function.
   Params:  entryP         Entry to clone.
   Returns: (MemHandle)    Clone of entryP or NULL if error.
*/
MemHandle
PrvCloneEntry (MemPtr entryP)
{
   // Declare local variables.
   MemHandle cloneH = (MemHandle) entryP;   // Return value.

   // Return value.
   return cloneH;
}

Err
PrvDeleteEntry (MemHandle entryH)
{
   // Return success.
   return 0;
}

#endif